import argparse
import datetime
import json
import re
from dataclasses import dataclass
from typing import Any, List, Dict, Optional, Callable, Union

from prompt_toolkit import print_formatted_text, HTML

# from keepersdk.authentication import keeper_auth
# from keepersdk.enterprise import enterprise_types
from ... import utils, crypto, api
from ...proto import pedm_pb2
from .. import base
from . import pedm_admin
from ..helpers import report_utils
from ...params import KeeperParams
from ...pedm import admin_plugin


class PedmReportCommand(base.GroupCommandNew):
    def __init__(self):
        super().__init__('Display PEDM  reports')
        self.register_command_new(PedmPolicyUsageReportCommand(), 'policy-usage', 'pu')
        self.register_command_new(PedmColumnReportCommand(), 'column', 'c')
        self.register_command_new(PedmEventReportCommand(), 'event', 'e')
        self.register_command_new(PedmEventSummaryReportCommand(), 'summary', 's')


@dataclass
class FieldInfo:
    name: str
    type: str
    protection: str

in_pattern = re.compile(r"\s*in\s*\(\s*(.*)\s*\)", re.IGNORECASE)
between_pattern = re.compile(r"\s*between\s+(\S*)\s+and\s+(.*)", re.IGNORECASE)
predefined_date_filters = {'today', 'yesterday', 'last_7_days', 'last_30_days', 'month_to_date', 'last_month',
                           'year_to_date', 'last_year'}

display_fields = ('deployment_uid', 'admin_uid', 'agent_uid', 'agent_status', 'agent_version',
                  'session_uid', 'session_type', 'policy_uid', 'policy_version',
                  'request_uid', 'evaluation_status', 'request_status', 'plugin_uid', 'plugin_version', 'update_status',
                  'user_info', 'target_info', 'reason')

class AuditMixin:
    syslog_templates: Optional[Dict[str, str]] = None
    field_info: Optional[Dict[str, FieldInfo]] = None
    user_lookup: Optional[Dict[int, str]] = None

    @staticmethod
    def load_audit_metadata(params) -> None:
        if AuditMixin.syslog_templates is None:
            rq = {
                'fields': ['audit_event_type', 'report_field']
            }
            rs = api.execute_router_json(params, 'pedm/get_audit_event_dimensions', rq)
            assert rs is not None
            AuditMixin.syslog_templates = {}
            AuditMixin.field_info = {}
            name: Optional[str]
            for et in rs['audit_event_type']:
                name = et.get('name')
                syslog = et.get('syslog')
                if name and syslog:
                    AuditMixin.syslog_templates[name] = syslog

            for rf in rs['report_field']:
                name = rf.get('name') or ''
                field_type = rf.get('type') or ''
                protection = rf.get('protection') or ''
                AuditMixin.field_info[name] = FieldInfo(name=name, type=field_type, protection=protection)

        if AuditMixin.user_lookup is None:
            AuditMixin.user_lookup = {}
            for u in params.enterprise['users']:
                AuditMixin.user_lookup[u['enterprise_user_id']] = u['username']

    @staticmethod
    def convert_date_filter(value: Any) -> Union[int, str]:
        if isinstance(value, datetime.datetime):
            value = value.timestamp()
        elif isinstance(value, datetime.date):
            dt = datetime.datetime.combine(value, datetime.datetime.min.time())
            value = dt.timestamp()
        elif isinstance(value, (int, float)):
            value = float(value)
        elif isinstance(value, str):
            if value in {predefined_date_filters}:
                return value
            if len(value) <= 10:
                value = datetime.datetime.strptime(value, '%Y-%m-%d')
            else:
                value = datetime.datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')
            value = value.timestamp()
        return int(value)

    @staticmethod
    def convert_str_or_int_filter(value: Any) -> Union[str, int]:
        if isinstance(value, str):
            if value.isdigit():
                return int(value)
            else:
                return value
        elif isinstance(value, int):
            return value
        return str(value)

    @staticmethod
    def get_filter(value: str, convert: Callable[[Any], Any]) -> Any:
        filter_value = value.strip()
        bet = between_pattern.match(filter_value)
        if bet is not None:
            dt1, dt2, *_ = bet.groups()
            dt1 = convert(dt1)
            dt2 = convert(dt2)
            return {'min': dt1, 'max': dt2}

        inp = in_pattern.match(filter_value)
        if inp is not None:
            arr = []
            for v in inp.groups()[0].split(','):
                arr.append(convert(v.strip()))
            return arr

        for prefix in ['>=', '<=', '>', '<', '=']:
            if filter_value.startswith(prefix):
                value = convert(filter_value[len(prefix):].strip())
                if prefix == '>=':
                    return {'min': value}
                if prefix == '<=':
                    return {'max': value}
                if prefix == '>':
                    return {'min': value, 'exclude_min': True}
                if prefix == '<':
                    return {'max': value, 'exclude_max': True}
                return value

        return convert(filter_value)

    @staticmethod
    def get_field_value(field: str, value: Any, *, report_type: str = 'raw') -> Any:
        if field in ('event_time', 'first_date', 'last_date'):
            if isinstance(value, str):
                return value
            if isinstance(value, (int, float)):
                value = int(value)
                dt = datetime.datetime.fromtimestamp(value, tz=datetime.timezone.utc)
                # dt = dt.replace(tzinfo=datetime.timezone.utc).astimezone(tz=None)
                if field == 'event_time':
                    if report_type in ('day', 'week'):
                        return dt.date()
                    if report_type == 'month':
                        return dt.strftime('%B, %Y')
                    if report_type == 'hour':
                        return dt.strftime('%Y-%m-%d @%H:00')
                return dt
        return value

    @staticmethod
    def get_enterprise_user_id(user_uid: str) -> Optional[int]:
        try:
            user_bytes = utils.base64_url_decode(user_uid)
            if len(user_bytes) == 16:
                return int.from_bytes(user_bytes[8:], byteorder='big')
        except:
            pass
        return None

    @staticmethod
    def get_enterprise_user_email(user_id: int) -> Optional[str]:
        if AuditMixin.user_lookup:
            return AuditMixin.user_lookup.get(user_id)
        return None


audit_column_description = '''
<b>Audit Column Report Command</b>
Returns unique values for audit report fields.

To get a list of all report fields:
Commander> <ansigreen>pedm report column report_field</ansigreen>

To get a list of all report events:
Commander> <ansigreen>pedm report column audit_event_type</ansigreen>
'''

class PedmColumnReportCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='report column', description='Run column data audit reports',
                                         parents=[base.report_output_parser])
        parser.add_argument('--syntax-help', dest='syntax_help', action='store_true', help='display help')
        parser.add_argument('column', nargs='?', help='Audit report column')

        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        if kwargs.get("syntax_help") is True:
            print_formatted_text(HTML(audit_column_description))
            return None

        AuditMixin.load_audit_metadata(context)
        assert AuditMixin.field_info is not None

        enterprise = context.enterprise
        tree_key = enterprise['unencrypted_tree_key']
        encrypted_ec_private_key = utils.base64_url_decode(enterprise['keys']['ecc_encrypted_private_key'])
        ec_private_key = crypto.load_ec_private_key(crypto.decrypt_aes_v2(encrypted_ec_private_key, tree_key))

        column = kwargs.get('column')
        if not (isinstance(column, str) and len(column) > 0):
            raise base.CommandError('"column" must be a non-empty string')

        f_info: Optional[FieldInfo] = None
        if column != "report_field":
            f_info = AuditMixin.field_info.get(column)
            if not f_info:
                raise base.CommandError(f'column "{column}" is not a known audit report column')
            if f_info.type != 'group':
                raise base.CommandError(f'column "{column}" is not a known audit report grouping column')
        rq = {
            'fields': [column]
        }
        rs = api.execute_router_json(context, 'pedm/get_audit_event_dimensions', rq)
        assert rs is not None
        rows: List[List[Any]] = []
        headers: List[str]
        if column == 'report_field':
            headers = ['name', 'type', 'protection']
        elif column == 'audit_event_type':
            headers = ['name', 'id', 'is_client', 'syslog']
        else:
            headers = ['value']

        dimension = rs.get(column)
        if not isinstance(dimension, list):
            raise base.CommandError('Server response is not supported')
        for d in dimension:
            if isinstance(d, dict):
                row = []
                for header in headers:
                    row.append(d.pop(header, None))
                if len(d) > 0:
                    for k, v in d.items():
                        headers.append(k)
                        row.append(v)
                rows.append(row)
        if f_info is not None and f_info.protection == 'hash':
            headers.insert(0, 'uid')
            uids = [x[0] for x in rows]
            uids = uids[:500]
            coll_rq = pedm_pb2.AuditCollectionRequest()
            # coll_rq.collectionName.append(f_info.name)
            coll_rq.valueUid.extend([utils.base64_url_decode(x) for x in uids])
            coll_rs =  api.execute_router(context,
                'pedm/get_audit_collections', coll_rq, rs_type=pedm_pb2.AuditCollectionResponse)
            value_lookup: Dict[str, str] = {}
            assert coll_rs is not None

            for v in coll_rs.values:
                value_uid = utils.base64_url_encode(v.valueUid)
                try:
                    decrypted_data = crypto.decrypt_ec(v.encryptedData, ec_private_key)
                    value_lookup[value_uid] = decrypted_data.decode('utf-8', 'ignore')
                except:
                    pass
            for row in rows:
                value = value_lookup.get(row[0])
                if value:
                    row.append(value)

        if kwargs.get('format') != 'json':
            headers = [report_utils.field_to_title(x) for x in headers]

        return report_utils.dump_report_data(rows, headers, fmt=kwargs.get('format'), filename=kwargs.get('output'),
                                             row_number=True, sort_by=0, column_width=80)

audit_report_description = '''
<b>Audit Event Report Command</b>

To get a list of event fields run the following command:
Commander> <ansigreen>pedm report column report_field</ansigreen>

Any field that has type <u>group</u> or <u>filter</u> can be used as filter

Filter syntax: <b>[FIELD_NAME]=[CRITERIA]</b>
where criteria is
1. single value: Example: "agent_uid=NJvK0I5RpuF0UFMwRKY_Dw"
2. list of values: Example: "agent_uid=IN(NJvK0I5RpuF0UFMwRKY_Dw, VYLhwqhRvhIpma9e1HoDFw)"
3. range value: Example: "created=BETWEEN 2024-01-01 AND 2024-02-01"
Predefined date range values: today, yesterday, last_7_days, last_30_days, month_to_date, last_month, year_to_date, last_year
<ansigreen>event_time=last_month</ansigreen>
'''
class PedmEventReportCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='report event', description='Run audit event reports',
                                         parents=[base.report_output_parser])
        parser.add_argument('--syntax-help', dest='syntax_help', action='store_true', help='display help')
        parser.add_argument('--report-format', dest='report_format', action='store', default='message',
                            choices=['message', 'fields'], help='output format (raw reports only)')
        parser.add_argument('--timezone', dest='timezone', action='store', help='return results for specific timezone')
        parser.add_argument('--limit', dest='limit', type=int, action='store',
                            help='maximum number of returned rows (1000 max)')
        parser.add_argument('--order', dest='order', action='store', choices=['desc', 'asc'], help='sort order')
        parser.add_argument('filter', nargs='*', help='Report filters')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        if kwargs.get("syntax_help") is True:
            print_formatted_text(HTML(audit_report_description))
            return

        AuditMixin.load_audit_metadata(context)
        assert AuditMixin.field_info is not None
        assert AuditMixin.syslog_templates is not None

        filters = kwargs.get('filter')
        if filters and isinstance(filters, str):
            filters = [filters]
        report_filter: Dict[str, Any] = {}
        for filter_arg in filters:
            field, sep, criteria = filter_arg.partition('=')
            if not sep:
                raise base.CommandError(f'Filter syntax error: {filter_arg}')
            info = AuditMixin.field_info.get(field)
            if not info:
                raise base.CommandError(f'field "{field}" is not a known audit report column')
            if info.type not in ('group', 'filter'):
                raise base.CommandError(f'column "{field}" is not a known audit report filter column')
            convert: Callable[[Any], Any]
            if filter_arg == 'event_time':
                convert = AuditMixin.convert_date_filter
            else:
                convert = AuditMixin.convert_str_or_int_filter
            report_filter[field] = AuditMixin.get_filter(criteria, convert)
        rq: Dict[str, Any] = {
            'timezone': datetime.datetime.now().astimezone().tzname()
        }
        if len(report_filter) > 0:
            rq['filter'] = report_filter
        limit = kwargs.get('limit')
        if limit is not None:
            rq['limit'] = limit
        order = kwargs.get('order')
        if order:
            rq['order'] = order
        rs = api.execute_router_json(context, 'pedm/get_audit_events', rq)
        assert rs is not None
        events = rs.get('audit_event_overview_report_rows')
        assert isinstance(events, list)
        for event in events:
            if 'admin_uid' in event:
                user_id = AuditMixin.get_enterprise_user_id(event['admin_uid'])
                if isinstance(user_id, int):
                    username = AuditMixin.get_enterprise_user_email(user_id)
                    if isinstance(username, str):
                        event['admin_uid'] = username

        if kwargs.get('format') == 'json':
            return json.dumps(events, indent=2)

        rows: List[List[Any]] = []
        headers: List[str] = []
        if kwargs.get('report_format') == 'message':
            headers.extend(('event_time', 'audit_event_type', 'message'))
            for event in events:
                event_type = event.get('audit_event_type')
                if not event_type:
                    rows.append([None, 'Event is missing "event_type" field'])
                    continue
                syslog = AuditMixin.syslog_templates.get(event_type)
                if not syslog:
                    rows.append([None, f'Syslog message is missing for event "{event_type}"'])
                    continue
                while True:
                    pattern = re.search(r'\${(\w+)}', syslog)
                    if pattern:
                        token = pattern[1]
                        value = event.get(token)
                        val = AuditMixin.get_field_value(token, value, report_type='raw')
                        if val is None:
                            val = '<missing>'
                        sp = pattern.span()
                        syslog = syslog[:sp[0]] + str(val) + syslog[sp[1]:]
                    else:
                        break
                event_time = event.get('event_time')
                e_time = AuditMixin.get_field_value('event_time', event_time, report_type='raw')
                rows.append([e_time, event_type, syslog])
        else:
            all_fields = set()
            if isinstance(events, list):
                for event in events:
                    all_fields.update(event.keys())
            headers = ['event_time', 'audit_event_type']
            headers.extend((x for x in display_fields if x in all_fields))
            for event in events:
                rows.append([AuditMixin.get_field_value(x, event.get(x)) for x in headers])

            headers = [report_utils.field_to_title(x) for x in headers]

        return report_utils.dump_report_data(rows, headers, fmt=kwargs.get('format'), filename=kwargs.get('output'),
                                             row_number=True)


audit_summary_report_description = '''
Audit Summary Report Command Syntax Description:

To get a list of event fields run the following command:
My Vault>  pedm report column report_field

Any field that has type "group" can be used as grouping column

Any field that has type "group" or "filter" can be used as filter

--group-by:             Defines break down report properties.

--aggregate:            Defines the aggregate value:
     occurrences        number of events. COUNT(*)
   first_created        starting date. MIN(created)
    last_created        ending date. MAX(created)

Filter syntax
<FIELD_NAME>=<CRITERIA>
where criteria is
1. single value: Example: "agent_uid=NJvK0I5RpuF0UFMwRKY_Dw"
2. list of values: Example: "agent_uid=IN(NJvK0I5RpuF0UFMwRKY_Dw, VYLhwqhRvhIpma9e1HoDFw)"
3. range value: Example: "created=BETWEEN 2024-01-01 AND 2024-02-01"
Predefined date range values: today, yesterday, last_7_days, last_30_days, month_to_date, last_month, year_to_date, last_year
"event_time=last_month"
'''

class PedmEventSummaryReportCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='report event', description='Run audit summary reports',
                                         parents=[base.report_output_parser])
        parser.add_argument('--syntax-help', dest='syntax_help', action='store_true', help='display help')
        parser.add_argument('--report-type', dest='report_type', action='store',
                            choices=['hour', 'day', 'month', 'span'], default='span', help='report type')
        parser.add_argument('--group-by', dest='group_by', action='append',
                            help='group by columns. (can be repeated).')
        parser.add_argument('--aggregate', dest='aggregate', action='append',
                            choices=['occurrences', 'first_date', 'last_date'],
                            help='aggregated value. (can be repeated).')
        parser.add_argument('--timezone', dest='timezone', action='store', help='return results for specific timezone')
        parser.add_argument('--limit', dest='limit', type=int, action='store',
                            help='maximum number of returned rows (2000 max)')
        parser.add_argument('--order', dest='order', action='store', choices=['desc', 'asc'], help='sort order')
        parser.add_argument('filter', nargs='*', help='Report filters')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        if kwargs.get("syntax_help") is True:
            return audit_summary_report_description

        AuditMixin.load_audit_metadata(context)
        assert AuditMixin.field_info is not None

        filters = kwargs.get('filter')
        if filters and isinstance(filters, str):
            filters = [filters]
        report_filter: Dict[str, Any] = {}
        for filter_arg in filters:
            field, sep, criteria = filter_arg.partition('=')
            if not sep:
                raise base.CommandError(f'Filter syntax error: {filter_arg}')
            info = AuditMixin.field_info.get(field)
            if not info:
                raise base.CommandError(f'field "{field}" is not a known audit report column')
            if info.type not in ('group', 'filter'):
                raise base.CommandError(f'column "{field}" is not a known audit report filter column')
            convert: Callable[[Any], Any]
            if filter_arg == 'event_time':
                convert = AuditMixin.convert_date_filter
            else:
                convert = AuditMixin.convert_str_or_int_filter
            report_filter[field] = AuditMixin.get_filter(criteria, convert)

        report_type: Optional[str] = kwargs.get('report_type')
        if not report_type:
            raise base.CommandError(f'"report-type" is a required argument')
        aggregate = kwargs.get('aggregate')
        if not aggregate:
            aggregate = ['occurrences']
        elif isinstance(aggregate, str):
            aggregate = [aggregate]

        rq: Dict[str, Any] = {
            'report_type': report_type,
            'aggregate': aggregate,
            'timezone': datetime.datetime.now().astimezone().tzname()
        }

        if len(report_filter) > 0:
            rq['filter'] = report_filter
        group_by = kwargs.get('group_by')
        if group_by:
            if isinstance(group_by, str):
                group_by = [group_by]
            rq['group_by'] = group_by

        limit = kwargs.get('limit') or 50
        rq['limit'] = limit
        order = kwargs.get('order') or 'desc'
        rq['order'] = order
        rs = api.execute_router_json(context, 'pedm/get_summary_audit_report', rq)
        assert rs is not None

        events = rs.get('audit_event_summary_report_rows')
        assert isinstance(events, list)

        if kwargs.get('format') == 'json':
            return json.dumps(events, indent=2)

        if not events:
            return

        headers = []
        if report_type != 'span':
            headers.append('event_time')
        headers.extend(aggregate)
        if group_by:
            headers.extend(group_by)
        rows: List[List[Any]] = []
        for event in events:
            rows.append([AuditMixin.get_field_value(x, event.get(x), report_type=report_type) for x in headers])

        headers = [report_utils.field_to_title(x) for x in headers]
        return report_utils.dump_report_data(
            rows, headers, fmt=kwargs.get('format'), filename=kwargs.get('output'), row_number=True)


class PedmPolicyUsageReportCommand(base.ArgparseCommand):
    def __init__(self):
        parser = argparse.ArgumentParser(prog='report policy-usage', description='Run audit summary reports',
                                         parents=[base.report_output_parser])
        parser.add_argument('--summary', dest='summary', action='store_true', help='Agent count only')
        parser.add_argument('policy', nargs='+', help='Policy UID')
        super().__init__(parser)

    def execute(self, context: KeeperParams, **kwargs) -> Any:
        plugin = admin_plugin.get_pedm_plugin(context)

        is_summary = kwargs.get('summary') is True
        rq = pedm_pb2.PolicyAgentRequest()
        rq.summaryOnly = is_summary
        policies = kwargs.get('policy')
        if not isinstance(policies, list):
            policies = [str(policies)]
        if '*' in policies:
            rq.policyUid.append(plugin.all_agents)
        else:
            policies = pedm_admin.PedmUtils.resolve_existing_policies(plugin, policies)
            if len(policies) == 0:
                raise base.CommandError(f'Cannot find any policy')
            rq.policyUid.extend([utils.base64_url_decode(x.policy_uid) for x in policies])

        rs = api.execute_router(context,'pedm/get_policy_agents', rq, rs_type=pedm_pb2.PolicyAgentResponse)
        assert rs is not None
        rows: List[List[Any]] = []
        headers: List[str]
        if is_summary:
            headers = ['policy_uid', 'agent_count']
            rows = [[[utils.base64_url_encode(x) for x in rq.policyUid], rs.agentCount]]
        else:
            headers = ['policy_uid', 'agent_uid']
            for i in range(max(len(rq.policyUid), len(rs.agentUid))):
                policy_uid = utils.base64_url_encode(rq.policyUid[i]) if i < len(rq.policyUid) else ''
                agent_uid = utils.base64_url_encode(rs.agentUid[i]) if i < len(rs.agentUid) else ''
                rows.append([policy_uid, agent_uid])

        return report_utils.dump_report_data(rows, headers, fmt=kwargs.get('format'), filename=kwargs.get('output'))


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Contact: ops@keepersecurity.com
#
import abc
import contextlib
import io
import json
import logging
import mimetypes
import os
import shutil
from typing import BinaryIO, Iterator, Optional, List, Union, Dict

import requests

from . import crypto, api, utils
from .params import KeeperParams
from .proto import record_pb2
from .record_facades import FileRefRecordFacade
from .vault import KeeperRecord, PasswordRecord, TypedRecord, FileRecord, AttachmentFile, AttachmentFileThumb


def prepare_attachment_download(params, record_uid, attachment_name=None):
    # type: (KeeperParams, str, Optional[str]) -> Iterator[AttachmentDownloadRequest]
    record = KeeperRecord.load(params, record_uid)
    if not record:
        logging.warning('Record UID \"%s\" not found.', record_uid)

    if record.version in {3, 4}:
        rq = record_pb2.FilesGetRequest()
        rq.for_thumbnails = False
        if isinstance(record, FileRecord):
            rq.record_uids.append(utils.base64_url_decode(record.record_uid))
        elif isinstance(record, TypedRecord):
            typed_field = record.get_typed_field('fileRef')
            if typed_field and isinstance(typed_field.value, list):
                for file_uid in typed_field.value:
                    file_record = KeeperRecord.load(params, file_uid)
                    if isinstance(file_record, FileRecord):
                        if attachment_name:
                            if attachment_name != file_uid and file_record.title.lower() != attachment_name.lower() and \
                                    file_record.name.lower() != attachment_name.lower():
                                continue
                        rq.record_uids.append(utils.base64_url_decode(file_uid))

        if len(rq.record_uids) > 0:
            rs = api.communicate_rest(params, rq, 'vault/files_download', rs_type=record_pb2.FilesGetResponse)
            for file_status in rs.files:
                file_uid = utils.base64_url_encode(file_status.record_uid)
                if file_status.status == record_pb2.FG_SUCCESS:
                    file_record = KeeperRecord.load(params, file_uid)
                    if isinstance(file_record, FileRecord):
                        adr = AttachmentDownloadRequest()
                        adr.file_id = file_uid
                        adr.url = file_status.url
                        adr.success_status_code = file_status.success_status_code
                        adr.encryption_key = file_record.record_key
                        adr.title = file_record.name
                        adr.is_gcm_encrypted = file_status.fileKeyType == record_pb2.ENCRYPTED_BY_DATA_KEY_GCM
                        yield adr
                else:
                    logging.warning('Error requesting download URL for file \"%s\"', file_uid)
    elif record.version == 2:
        attachments = []   # type: List[AttachmentFile]
        if isinstance(record, PasswordRecord):
            for atta in (record.attachments or []):
                if attachment_name:
                    if attachment_name != atta.id and attachment_name.lower() != atta.title.lower() and \
                            attachment_name.lower() != atta.name.lower():
                        continue
                attachments.append(atta)
        if len(attachments) > 0:
            rq = {
                'command': 'request_download',
                'file_ids': [x.id for x in attachments],
            }
            api.resolve_record_access_path(params, record_uid, path=rq)
            rs = api.communicate(params, rq)

            if rs['result'] == 'success':
                for attachment, dl in zip(attachments, rs['downloads']):
                    if 'url' in dl:
                        adr = AttachmentDownloadRequest()
                        adr.file_id = attachment.id
                        adr.title = attachment.name
                        adr.url = dl['url']
                        adr.encryption_key = utils.base64_url_decode(attachment.key)
                        adr.is_gcm_encrypted = False
                        yield adr


class AttachmentDownloadRequest:
    def __init__(self):
        self.file_id = ''
        self.url = ''
        self.encryption_key = b''
        self.title = ''
        self.is_gcm_encrypted = False
        self.success_status_code = 200

    def download_to_file(self, params, file_name):  # type: (KeeperParams, str) -> None
        logging.info('Downloading \'%s\'', os.path.abspath(file_name))
        with open(file_name, 'wb') as file_stream:
            self.download_to_stream(params, file_stream)

    def download_to_stream(self, params, output_stream):  # type: (KeeperParams, BinaryIO) -> int
        with requests.get(self.url, proxies=params.rest_context.proxies, stream=True) as rq_http:
            if self.success_status_code != rq_http.status_code:
                logging.warning('HTTP status code: %d', rq_http.status_code)
            crypter = crypto.StreamCrypter()
            crypter.is_gcm = self.is_gcm_encrypted
            crypter.key = self.encryption_key
            with crypter.set_stream(rq_http.raw, for_encrypt=False) as attachment:
                shutil.copyfileobj(attachment, output_stream, 10240)
            output_stream.flush()
            return crypter.bytes_read


class UploadTask(abc.ABC):
    def __init__(self):
        self.mime_type = ''
        self.size = 0
        self.name = ''
        self.title = ''
        self.thumbnail = None   # type: Optional[bytes]

    def prepare(self):
        pass

    @abc.abstractmethod
    def open(self):  # type: () -> BinaryIO
        pass


class BytesUploadTask(UploadTask):
    def __init__(self, data):
        super().__init__()
        self.data = data if isinstance(data, bytes) else b''
        self.size = len(self.data)

    @contextlib.contextmanager
    def open(self):
        yield io.BytesIO(self.data)


class FileUploadTask(UploadTask):
    def __init__(self, file_path):
        super().__init__()
        self.file_path = file_path
        self.name = os.path.basename(self.file_path)

    def prepare(self):
        self.file_path = os.path.expanduser(self.file_path)
        if not os.path.isfile(self.file_path):
            raise ValueError(f'File {self.file_path} does not exist')
        self.size = os.path.getsize(self.file_path)
        if not self.mime_type:
            mt = mimetypes.guess_type(self.file_path)
            if isinstance(mt, tuple) and mt[0]:
                self.mime_type = mt[0]

    @contextlib.contextmanager
    def open(self):
        yield open(self.file_path, 'rb')


def upload_attachments(params, record, attachments):
    # type: (KeeperParams, Union[PasswordRecord, TypedRecord], List[UploadTask]) -> None
    cryptor = crypto.StreamCrypter()
    if isinstance(record, PasswordRecord):
        cryptor.is_gcm = False
        if not isinstance(record.attachments, list):
            record.attachments = []
        thumbs = [x for x in attachments if x.thumbnail is not None]
        rq = {
            'command': 'request_upload',
            'file_count': len(attachments),
            'thumbnail_count': len(thumbs),
        }
        rs = api.communicate(params, rq)
        file_uploads = rs['file_uploads']
        thumb_uploads = rs['thumbnail_uploads']
        thumb_pos = 0
        for i, task in enumerate(attachments):
            uo = file_uploads[i]
            attachment_id = uo['file_id']
            attachment_key = utils.generate_aes_key()
            cryptor.key = attachment_key
            atta = AttachmentFile()
            task.prepare()
            with task.open() as task_stream, cryptor.set_stream(task_stream, True) as crypto_stream:
                files = {
                    uo['file_parameter']: (attachment_id, crypto_stream, 'application/octet-stream')
                }
                response = requests.post(uo['url'], files=files, data=uo['parameters'])
                if response.status_code == uo['success_status_code']:
                    atta.id = attachment_id
                    atta.name = task.name or ''
                    atta.title = task.title or ''
                    atta.mime_type = task.mime_type or ''
                    atta.last_modified = utils.current_milli_time()
                    atta.key = utils.base64_url_encode(attachment_key)
                    atta.size = task.size
                    record.attachments.append(atta)
                else:
                    raise Exception(f'Uploading file {task.name}: HTTP status code {response.status_code}')
            if isinstance(task.thumbnail, bytes) and thumb_pos < len(thumbs) and thumb_pos < len(thumb_uploads):
                thumb_task = thumbs[thumb_pos]
                tuo = thumb_uploads[thumb_pos]
                thumb_pos += 1
                atta.thumbnails = []
                with io.BytesIO(task.thumbnail) as thumb_stream, \
                        cryptor.set_stream(thumb_stream, True) as crypto_stream:
                    files = {
                        tuo['file_parameter']: (tuo['file_id'], crypto_stream, 'application/octet-stream')
                    }
                    response = requests.post(tuo['url'], files=files, data=tuo['parameters'])
                    if response.status_code == uo['success_status_code']:
                        thumb = AttachmentFileThumb()
                        thumb.id = tuo['file_id']
                        thumb.type = thumb_task.mime_type
                        thumb.size = len(task.thumbnail)
                        atta.thumbnails.append(thumb)
                    else:
                        logging.warning(
                            'Uploading thumbnail %s: HTTP status code %d', task.name, response.status_code)

    elif isinstance(record, TypedRecord):
        cryptor.is_gcm = True
        rq = record_pb2.FilesAddRequest()
        rq.client_time = utils.current_milli_time()
        file_keys = {}   # type: Dict[bytes, bytes]
        file_tasks = {}   # type: Dict[bytes, UploadTask]
        for task in attachments:
            task.prepare()
            file_uid = utils.base64_url_decode(utils.generate_uid())
            file_key = utils.generate_aes_key()
            file_keys[file_uid] = file_key
            file_tasks[file_uid] = task
            file = record_pb2.File()
            file.record_uid = file_uid
            file.record_key = crypto.encrypt_aes_v2(file_key, params.data_key)
            file.fileSize = task.size + 100
            file_data = {
                'title': task.title or task.name,
                'name': task.name or '',
                'type': task.mime_type or '',
                'size': task.size
            }
            if isinstance(task.thumbnail, bytes):
                file_data['thumbnail_size'] = len(task.thumbnail)
                file.thumbSize = len(task.thumbnail) + 100
            file.data = crypto.encrypt_aes_v2(json.dumps(file_data).encode(), file_key)
            rq.files.append(file)

        facade = FileRefRecordFacade()
        facade.record = record

        rs = api.communicate_rest(params, rq, 'vault/files_add', rs_type=record_pb2.FilesAddResponse)
        for uo in rs.files:
            file_uid = uo.record_uid
            task = file_tasks[file_uid]
            if uo.status != record_pb2.FA_SUCCESS:
                raise Exception(f'Uploading file {task.name}: Get upload URL error.')

            file_key = file_keys[file_uid]
            cryptor.key = file_key
            with task.open() as task_stream, cryptor.set_stream(task_stream, True) as crypto_stream:
                file_ref = utils.base64_url_encode(file_uid)
                files = {
                    'file': (file_ref, crypto_stream, 'application/octet-stream')
                }
                response = requests.post(uo.url, files=files, data=json.loads(uo.parameters),
                                         proxies=params.rest_context.proxies,
                                         verify=params.rest_context.certificate_check)
                if response.status_code == uo.success_status_code:
                    facade.file_ref.append(file_ref)
                    if record.linked_keys is None:
                        record.linked_keys = {}
                    record.linked_keys[file_ref] = file_key
                else:
                    raise Exception(f'Uploading file {task.name}: HTTP status code {response.status_code}')
            if isinstance(task.thumbnail, bytes):
                try:
                    with io.BytesIO(task.thumbnail) as thumb_stream, \
                            cryptor.set_stream(thumb_stream, True) as crypto_stream:
                        files = {
                            'thumb': crypto_stream
                        }
                        requests.post(uo.url, files=files, data=json.loads(uo.thumbnail_parameters),
                                      proxies=params.rest_context.proxies,
                                      verify=params.rest_context.certificate_check)
                except Exception as e:
                    logging.warning('Error uploading thumbnail: %s', e)
    else:
        raise Exception(f'Unsupported record type: {type(record)}')



#  _  __  
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|            
#
# Keeper Commander 
# Contact: ops@keepersecurity.com
#

import abc
import hashlib
import logging
import os
import secrets
import string
from secrets import choice
from typing import Optional, List, Iterator
from collections import namedtuple

from . import crypto

from Cryptodome.Random.random import shuffle

DEFAULT_PASSWORD_LENGTH = 20
PW_SPECIAL_CHARACTERS = '!@#$%()+;<>=?[]{}^.,'

PasswordStrength = namedtuple('PasswordStrength', 'length caps lower digits symbols')


def get_password_strength(password):  # type: (str) -> PasswordStrength
    length = len(password)
    caps = 0
    lower = 0
    digits = 0
    symbols = 0

    for ch in password:
        if ch.isalpha():
            if ch.isupper():
                caps += 1
            else:
                lower += 1
        elif ch.isdigit():
            digits += 1
        elif ch in PW_SPECIAL_CHARACTERS:
            symbols += 1
    return PasswordStrength(length=length, caps=caps, lower=lower, digits=digits, symbols=symbols)


def generate(length=64):
    generator = KeeperPasswordGenerator(length=length)
    return generator.generate()


class PasswordGenerator(abc.ABC):
    @abc.abstractmethod
    def generate(self):   # type: () -> str
        pass


class KeeperPasswordGenerator(PasswordGenerator):
    def __init__(self, length: int = DEFAULT_PASSWORD_LENGTH,
                 symbols: Optional[int] = None,
                 digits: Optional[int] = None,
                 caps: Optional[int] = None,
                 lower: Optional[int] = None,
                 special_characters: str = PW_SPECIAL_CHARACTERS):

        sum_categories = sum((abs(i) if isinstance(i, int) else 0) for i in (symbols, digits, caps, lower))
        extra_count = length - sum_categories if length > sum_categories else 0
        extra_chars = ''
        if symbols is None or isinstance(symbols, int) and symbols > 0:
            extra_chars += special_characters
        if digits is None or isinstance(digits, int) and digits > 0:
            extra_chars += string.digits
        if caps is None or isinstance(caps, int) and caps > 0:
            extra_chars += string.ascii_uppercase
        if lower is None or isinstance(lower, int) and lower > 0:
            extra_chars += string.ascii_lowercase
        if extra_count > 0 and not extra_chars:
            if isinstance(symbols, int) and symbols < 0:
                extra_chars += special_characters
            if isinstance(digits, int) and digits < 0:
                extra_chars += string.digits
            if isinstance(caps, int) and caps < 0:
                extra_chars += string.ascii_uppercase
            if isinstance(lower, int) and lower < 0:
                extra_chars += string.ascii_lowercase

            if extra_count > 0 and not extra_chars:
                raise Exception('Password character set is empty')
        self.category_map = [
            (abs(symbols) if isinstance(symbols, int) else 0, special_characters),
            (abs(digits) if isinstance(digits, int) else 0, string.digits),
            (abs(caps) if isinstance(caps, int) else 0, string.ascii_uppercase),
            (abs(lower) if isinstance(lower, int) else 0, string.ascii_lowercase),
            (extra_count, extra_chars)
        ]

    def generate(self) -> str:
        password_list = []
        for count, chars in self.category_map:
            password_list.extend(choice(chars) for i in range(count))
        shuffle(password_list)
        return ''.join(password_list)

    @classmethod
    def create_from_rules(cls, rule_string: str, length: Optional[int] = None,
                          special_characters: str = PW_SPECIAL_CHARACTERS):
        """Create instance of class from rules string

        rule_string: comma separated integer character counts of [length,] uppercase, lowercase, numbers, symbols
        length: length of password
        special_characters: set of characters used to generate password symbols
        """
        try:
            rule_list = [int(s.strip()) for s in rule_string.split(',')]
            if len(rule_list) == 5:
                l = rule_list.pop(0)
                if not length:
                    length = l
            if len(rule_list) != 4:
                raise Exception('Invalid rules')
        except:
            logging.warning('Invalid rules to generate password. Format is "[length,] upper, lower, digits, symbols"')
            return None

        if length is None:
            length = DEFAULT_PASSWORD_LENGTH
        upper, lower, digits, symbols = rule_list
        length = sum(rule_list) if length is None else length
        return cls(length=length, caps=upper, lower=lower, digits=digits, symbols=symbols, special_characters=special_characters)


class DicewarePasswordGenerator(PasswordGenerator):
    def __init__(self, number_of_rolls, word_list_file=None, delimiter=' '):   # type: (int, Optional[str], str) -> None
        self._number_of_rolls = number_of_rolls if number_of_rolls > 0 else 5
        self.delimiter = delimiter

        if word_list_file:
            dice_path = os.path.join(os.path.dirname(__file__), 'resources', word_list_file)
            if not os.path.isfile(dice_path):
                dice_path = os.path.expanduser(word_list_file)
        else:
            dice_path = os.path.join(os.path.dirname(__file__), 'resources', 'diceware.wordlist.asc.txt')
        self._vocabulary = None    # type: Optional[List[str]]
        if os.path.isfile(dice_path):
            with open(dice_path, 'r', encoding='utf-8') as dw:
                self._vocabulary = []
                line_count = 0
                unique_words = set()
                for line in dw.readlines():
                    if not line:
                        continue
                    if line.startswith('--'):
                        continue
                    line_count += 1
                    words = [x.strip() for x in line.split()]
                    word = words[1] if len(words) >= 2 else words[0]
                    self._vocabulary.append(word)
                    unique_words.add(word.lower())
                if line_count != len(unique_words):
                    raise Exception(f'Word list file \"{dice_path}\" contains non-unique words.')
        else:
            raise Exception(f'Word list file \"{dice_path}\" not found.')

    def generate(self):
        if not self._vocabulary:
            raise Exception(f'Diceware word list was not loaded')

        words = [secrets.choice(self._vocabulary) for _ in range(self._number_of_rolls)]
        shuffle(words)
        return self.delimiter.join(words)


class CryptoPassphraseGenerator(PasswordGenerator):
    def __init__(self):
        self._vocabulary = None    # type: Optional[List[str]]
        dice_path = os.path.join(os.path.dirname(__file__), 'resources', 'bip-39.english.txt')
        if os.path.isfile(dice_path):
            with open(dice_path, 'r', encoding='utf-8') as dw:
                self._vocabulary = []
                for line in dw.readlines():
                    if not line:
                        continue
                    if line.startswith('--'):
                        continue
                    words = [x.strip() for x in line.split()]
                    word = words[1] if len(words) >= 2 else words[0]
                    self._vocabulary.append(word)

                unique_words = set((x.lower() for x in self._vocabulary))
                if len(self._vocabulary) != len(unique_words):
                    raise Exception(f'Word list file \"{dice_path}\" contains non-unique words.')
                if len(unique_words) != 2 ** 11:
                    raise Exception(f'Word list file \"{dice_path}\" is incorrect crypto dictionary.')
        else:
            raise Exception(f'Word list file \"{dice_path}\" not found.')

    def get_vocabulary(self):    # type: () -> Iterator[str]
        return (x for x in self._vocabulary)

    def generate(self):
        key = crypto.get_random_bytes(32)
        hasher = hashlib.sha256()
        hasher.update(key)
        digest = hasher.digest()
        secret = int.from_bytes(key + digest[:1], byteorder='big')

        words = []
        for i in range(24):
            words.append(secret & 0x07ff)
            secret >>= 11

        words.reverse()
        return ' '.join((self._vocabulary[x] for x in words))



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2024 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
from typing import Any, Dict, Optional
from ..config.service_config import ServiceConfig
from ..config.config_validation import ValidationError
from ..decorators.logging import logger, debug_decorator
from ...params import KeeperParams
from ...commands.base import report_output_parser, Command
from dataclasses import dataclass

@dataclass
class StreamlineArgs:
    port: Optional[int]
    allowedip: Optional[str]
    deniedip: Optional[str]
    commands: Optional[str]
    ngrok: Optional[str]
    ngrok_custom_domain: Optional[str]
    cloudflare: Optional[str]
    cloudflare_custom_domain: Optional[str]
    certfile: Optional[str]
    certpassword : Optional[str]
    fileformat : Optional[str]
    run_mode: Optional[str]
    queue_enabled: Optional[str]
    
class CreateService(Command):
    """Command to create a new service configuration."""
    
    def __init__(self):
        self.service_config = ServiceConfig()
        self._config_handler = None
        self._security_handler = None

    @property
    def config_handler(self):
        if self._config_handler is None:
            from .service_config_handlers import ServiceConfigHandler
            self._config_handler = ServiceConfigHandler(self.service_config)
        return self._config_handler

    @property
    def security_handler(self):
        if self._security_handler is None:
            from .security_config_handler import SecurityConfigHandler
            self._security_handler = SecurityConfigHandler(self.service_config)
        return self._security_handler

    @debug_decorator
    def get_parser(self):
        parser = argparse.ArgumentParser(prog='service-create', parents=[report_output_parser], description='Creates and initializes the Commander REST API service')
        parser.add_argument('-p', '--port', type=int, help='port number for the service (required)')
        parser.add_argument('-aip', '--allowedip', type=str, help='allowed ip to access service')
        parser.add_argument('-dip', '--deniedip', type=str, help='denied ip to access service')
        parser.add_argument('-c', '--commands', type=str, help='command list for policy')
        parser.add_argument('-ng', '--ngrok', type=str, help='ngrok auth token to generate public URL (optional)')
        parser.add_argument('-cd', '--ngrok_custom_domain', type=str, help='ngrok custom domain name(optional)')
        parser.add_argument('-cf', '--cloudflare', type=str, help='cloudflare tunnel token to generate public URL (required when using cloudflare)')
        parser.add_argument('-cfd', '--cloudflare_custom_domain', type=str, help='cloudflare custom domain name (required when using cloudflare)')
        parser.add_argument('-crtf', '--certfile', type=str, help='certificate file path')
        parser.add_argument('-crtp', '--certpassword', type=str, help='certificate password')
        parser.add_argument('-f', '--fileformat', type=str, help='file format')
        parser.add_argument('-rm', '--run_mode', type=str, help='run mode')
        parser.add_argument('-q', '--queue_enabled', type=str, help='enable request queue (y/n)')
        return parser
    
    def execute(self, params: KeeperParams, **kwargs) -> None:
        try:
            from ..core.service_manager import ServiceManager
            if ServiceManager.get_status().startswith("Commander Service is Running"):
                print("Error: Commander Service is already running.")
                return

            from ..core.globals import init_globals
            init_globals(params)

            config_data = self.service_config.create_default_config()

            filtered_kwargs = {k: v for k, v in kwargs.items() if k in ['port', 'allowedip', 'deniedip', 'commands', 'ngrok', 'ngrok_custom_domain', 'cloudflare', 'cloudflare_custom_domain', 'certfile', 'certpassword', 'fileformat', 'run_mode', 'queue_enabled']}
            args = StreamlineArgs(**filtered_kwargs)
            self._handle_configuration(config_data, params, args)
            self._create_and_save_record(config_data, params, args)
            self._upload_and_start_service(params)

        except ValidationError as e:
            print(f"Validation Error: {str(e)}")
        except Exception as e:
            print(f"Unexpected error: {str(e)}")
    
    @debug_decorator
    def _handle_configuration(self, config_data: Dict[str, Any], params: KeeperParams, args: StreamlineArgs) -> None:       
        if args.port is not None:
            logger.debug("Entering streamlined configuration")
            self.config_handler.handle_streamlined_config(config_data, args, params)
        else:
            logger.debug("Entering interactive configuration")
            self.config_handler.handle_interactive_config(config_data, params)
            self.security_handler.configure_security(config_data)
    
    def _create_and_save_record(self, config_data: Dict[str, Any], params: KeeperParams, args: StreamlineArgs) -> None:
        if args.port is None:
            self.config_handler._configure_run_mode(config_data)
        
        record = self.service_config.create_record(config_data["is_advanced_security_enabled"], params, args.commands)
        config_data["records"] = [record]
        if config_data.get("fileformat"):
            format_type = config_data["fileformat"]
        else:
            format_type = self.service_config.format_handler.get_config_format('create')
            config_data["fileformat"] = format_type
        self.service_config.save_config(config_data, format_type)
        if config_data.get("tls_certificate") == "y":
            self.service_config.save_cert_data(config_data, 'create')
        
    def _upload_and_start_service(self, params: KeeperParams) -> None:
        self.service_config.update_or_add_record(params)
        from ..core.service_manager import ServiceManager
        ServiceManager.start_service()


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
from typing import Dict, Any, Tuple

from .... import utils, crypto
from ...utils.constants import (
    AUTH_REASONS,
    ERROR_MESSAGES,
    DEFAULT_BIOMETRIC_TIMEOUT
)
from ...utils.error_handler import BiometricErrorHandler

# WebAuthn constants
WEBAUTHN_CHALLENGE_TYPE_CREATE = 'webauthn.create'
WEBAUTHN_CHALLENGE_TYPE_GET = 'webauthn.get'
WEBAUTHN_CREDENTIAL_TYPE = 'public-key'

# Cryptographic constants
EC_CURVE_P256 = 1
EC_ALG_ES256 = -7
EC_KTY_EC2 = 2
EC_PUBLIC_KEY_UNCOMPRESSED_PREFIX = 0x04
EC_COORDINATE_LENGTH = 32

# Authentication flags
AUTH_FLAG_UP = 0b00000001  # User Present
AUTH_FLAG_UV = 0b00000100  # User Verified
AUTH_FLAG_AT = 0b01000000  # Attested credential data included
AUTH_FLAG_CREATION = AUTH_FLAG_UP | AUTH_FLAG_UV | AUTH_FLAG_AT  # 0b01000101
AUTH_FLAG_ASSERTION = AUTH_FLAG_UP | AUTH_FLAG_UV  # 0b00000101


class BaseWebAuthnClient:
    """Base WebAuthn client with common functionality"""
    
    def __init__(self, client_data_collector, keychain_manager):
        self.client_data_collector = client_data_collector
        self.keychain_manager = keychain_manager

    def _validate_dependencies(self, required_modules: list) -> None:
        """Validate that required modules are available"""
        BiometricErrorHandler.validate_dependencies(required_modules)

    def _create_client_data(self, data_type: str, challenge: bytes, rp_id: str) -> Tuple[Dict[str, Any], bytes]:
        """Create WebAuthn client data"""
        origin = getattr(self.client_data_collector, 'origin', f'https://{rp_id}')
        client_data = {
            'type': data_type,
            'challenge': utils.base64_url_encode(challenge),
            'origin': origin,
            'crossOrigin': False
        }
        client_data_json = json.dumps(client_data, separators=(',', ':')).encode()
        return client_data, client_data_json

    def _create_authenticator_data(self, rp_id: str, flags: int, counter: int = 0, 
                                 additional_data: bytes = b'') -> bytes:
        """Create WebAuthn authenticator data"""
        import hashlib
        import struct
        
        rp_id_hash = hashlib.sha256(rp_id.encode()).digest()
        counter_bytes = struct.pack('>I', counter)
        return rp_id_hash + struct.pack('B', flags) + counter_bytes + additional_data

    def _create_signed_data(self, authenticator_data: bytes, client_data_json: bytes) -> bytes:
        """Create data to be signed for WebAuthn"""
        import hashlib
        client_data_hash = hashlib.sha256(client_data_json).digest()
        return authenticator_data + client_data_hash


class MacOSTouchIDWebAuthnClient(BaseWebAuthnClient):
    """macOS Touch ID WebAuthn client with DRY principles"""
    
    def __init__(self, client_data_collector, keychain_manager):
        super().__init__(client_data_collector, keychain_manager)

    def make_credential(self, options):
        """Create WebAuthn credential using Touch ID"""
        try:
            self._validate_dependencies(['LocalAuthentication', 'cbor2'])
            
            timeout_ms = getattr(options, 'timeout', None)
            timeout_seconds = (timeout_ms / 1000.0) if timeout_ms else DEFAULT_BIOMETRIC_TIMEOUT
            
            if hasattr(options, 'exclude_credentials') and options.exclude_credentials:
                for excluded_cred in options.exclude_credentials:
                    cred_id = excluded_cred.id
                    if isinstance(cred_id, str):
                        cred_id_b64 = cred_id
                    else:
                        cred_id_b64 = utils.base64_url_encode(cred_id)
                    
                    if self.keychain_manager.credential_exists(cred_id_b64, options.rp.id, timeout_seconds):
                        raise OSError("The object already exists")
            
            credential_id = utils.base64_url_encode(crypto.get_random_bytes(32))
            private_key, public_key = crypto.generate_ec_key()
            public_key_data = crypto.unload_ec_public_key(public_key)
            private_key_data = crypto.unload_ec_private_key(private_key)

            rp_id = options.rp.id
            if not rp_id:
                raise Exception("No RP ID found in options - server configuration error")
            context = self._create_auth_context()
            self._check_biometric_availability(context)
            
            if not self.keychain_manager.store_credential(credential_id, private_key_data, rp_id, timeout_seconds):
                raise Exception(ERROR_MESSAGES['keychain_store_failed'])
            
            reason = AUTH_REASONS['register'].format(rp_id=rp_id)
            success = self._authenticate_with_touchid(context, reason, timeout_seconds)
            
            if not success:
                self.keychain_manager.delete_credential(credential_id, rp_id, timeout_seconds)
                raise Exception(ERROR_MESSAGES['authentication_failed'])

            return self._create_credential_response(
                options, credential_id, public_key_data, rp_id
            )

        except Exception as e:
            raise Exception(str(e))

    def get_assertion(self, options):
        """Get WebAuthn assertion using Touch ID"""
        try:
            self._validate_dependencies(['LocalAuthentication'])
            
            timeout_ms = getattr(options, 'timeout', None)
            timeout_seconds = (timeout_ms / 1000.0) if timeout_ms else DEFAULT_BIOMETRIC_TIMEOUT
            
            challenge = options.challenge
            rp_id = options.rp_id
            if not rp_id:
                raise Exception("No RP ID found in options - server configuration error")
            private_key, credential_id_b64, credential_id_bytes = self._find_credential(options, rp_id, timeout_seconds)
            
            context = self._create_auth_context()
            self._check_biometric_availability(context)
            
            reason = AUTH_REASONS['login'].format(rp_id=rp_id)
            success = self._authenticate_with_touchid(context, reason, timeout_seconds)
            
            if not success:
                raise Exception(ERROR_MESSAGES['authentication_failed'])

            return self._create_assertion_response(
                challenge, rp_id, credential_id_b64, credential_id_bytes, private_key
            )

        except Exception as e:
            raise Exception(str(e))

    def _create_auth_context(self):
        """Create LocalAuthentication context"""
        import LocalAuthentication  # pylint: disable=import-error
        return LocalAuthentication.LAContext.alloc().init()  # pylint: disable=no-member

    def _check_biometric_availability(self, context):
        """Check if biometric authentication is available"""
        import LocalAuthentication  # pylint: disable=import-error
        
        error = None
        can_evaluate = context.canEvaluatePolicy_error_(
            LocalAuthentication.LAPolicyDeviceOwnerAuthenticationWithBiometrics,  # pylint: disable=no-member
            error
        )
        
        if not can_evaluate:
            raise Exception(ERROR_MESSAGES['touchid_not_available'])

    def _authenticate_with_touchid(self, context, reason: str, timeout_seconds: float = 30) -> bool:
        """Perform Touch ID authentication with proper synchronization"""
        import LocalAuthentication  # pylint: disable=import-error
        import threading
                
        event = threading.Event()
        result = {'success': False, 'error': None}
        
        def auth_callback(success, error):
            result['success'] = bool(success)
            result['error'] = error
            event.set()
        
        context.evaluatePolicy_localizedReason_reply_(
            LocalAuthentication.LAPolicyDeviceOwnerAuthenticationWithBiometrics,  # pylint: disable=no-member
            reason,
            auth_callback
        )
        
        # Wait for result with synchronization
        if event.wait(timeout=timeout_seconds):
            if result['error']:
                from ...utils.error_handler import BiometricErrorHandler
                raise BiometricErrorHandler.handle_authentication_error(
                    Exception(str(result['error'])), "Touch ID"
                )
            return result['success']
        else:
            raise TimeoutError("Touch ID authentication timed out")

    def _find_credential(self, options, rp_id: str, timeout_seconds: float) -> Tuple[Any, str, bytes]:
        """Find credential using keychain manager"""
        allowed_credentials = options.allow_credentials or []
        if not allowed_credentials:
            raise Exception("No allowed credentials found")

        for cred in allowed_credentials:
            cred_id = cred.id
            if isinstance(cred_id, str):
                test_id_b64 = cred_id
                test_id_bytes = utils.base64_url_decode(cred_id)
            else:
                test_id_bytes = cred_id
                test_id_b64 = utils.base64_url_encode(cred_id)

            test_key = self.keychain_manager.load_credential(test_id_b64, rp_id, timeout_seconds)
            if test_key:
                return test_key, test_id_b64, test_id_bytes

        from ...utils.error_handler import BiometricErrorHandler
        raise BiometricErrorHandler.handle_authentication_error(
            Exception("no matching credential found"), "Touch ID"
        )

    def _create_credential_response(self, options, credential_id: str, 
                                  public_key_data: bytes, rp_id: str):
        """Create WebAuthn credential response"""
        import cbor2
        import struct
        
        challenge = options.challenge
        client_data, client_data_json = self._create_client_data(
            WEBAUTHN_CHALLENGE_TYPE_CREATE, challenge, rp_id
        )

        # Create COSE key
        x_coord, y_coord = self._extract_key_coordinates(public_key_data)
        cose_key = {
            1: EC_KTY_EC2,      # kty: EC2
            3: EC_ALG_ES256,    # alg: ES256
            -1: EC_CURVE_P256,  # crv: P-256
            -2: x_coord,
            -3: y_coord
        }

        # Create attested credential data
        attested_credential_data = (
            b'\x00' * 16 +  # AAGUID
            struct.pack('>H', len(utils.base64_url_decode(credential_id))) +
            utils.base64_url_decode(credential_id) +
            cbor2.dumps(cose_key)
        )

        # Create authenticator data
        authenticator_data = self._create_authenticator_data(
            rp_id, AUTH_FLAG_CREATION, 0, attested_credential_data
        )

        # Create attestation object
        attestation_object = {
            'fmt': 'none',
            'attStmt': {},
            'authData': authenticator_data
        }

        return self._create_registration_response(
            credential_id,
            client_data_json,
            cbor2.dumps(attestation_object)
        )

    def _create_assertion_response(self, challenge: bytes, rp_id: str, 
                                 credential_id_b64: str, credential_id_bytes: bytes, 
                                 private_key):
        """Create WebAuthn assertion response"""
        from cryptography.hazmat.primitives import hashes
        from cryptography.hazmat.primitives.asymmetric import ec
        
        client_data, client_data_json = self._create_client_data(
            WEBAUTHN_CHALLENGE_TYPE_GET, challenge, rp_id
        )

        # Create authenticator data
        authenticator_data = self._create_authenticator_data(rp_id, AUTH_FLAG_ASSERTION)

        # Sign the data
        signed_data = self._create_signed_data(authenticator_data, client_data_json)
        der_signature = private_key.sign(signed_data, ec.ECDSA(hashes.SHA256()))

        return self._create_authentication_response(
            credential_id_b64,
            credential_id_bytes,
            client_data_json,
            authenticator_data,
            der_signature
        )

    def _extract_key_coordinates(self, public_key_data: bytes) -> Tuple[bytes, bytes]:
        """Extract x and y coordinates from EC key"""
        expected_length = 1 + (2 * EC_COORDINATE_LENGTH)  # 1 + (2 * 32) = 65
        if len(public_key_data) != expected_length or public_key_data[0] != EC_PUBLIC_KEY_UNCOMPRESSED_PREFIX:
            raise Exception("Invalid P-256 public key format")
        
        return (public_key_data[1:1 + EC_COORDINATE_LENGTH], 
                public_key_data[1 + EC_COORDINATE_LENGTH:])

    def _create_registration_response(self, credential_id: str, client_data_json: bytes, 
                                    attestation_object_cbor: bytes):
        """Create registration response"""
        class RegistrationResponse:
            def __init__(self, cred_id, cred_raw_id, client_data, attestation_obj):
                self.id = cred_id
                self.raw_id = cred_raw_id
                self.response = AttestationResponse(client_data, attestation_obj)
                self.client_extension_results = {}
                self.type = WEBAUTHN_CREDENTIAL_TYPE
                
        class AttestationResponse:
            def __init__(self, client_data, attestation_obj):
                self.client_data = client_data
                self.attestation_object = attestation_obj
        
        return RegistrationResponse(
            credential_id,
            utils.base64_url_decode(credential_id),
            client_data_json,
            attestation_object_cbor
        )

    def _create_authentication_response(self, credential_id_b64: str, credential_id_bytes: bytes, 
                                      client_data_json: bytes, authenticator_data: bytes, 
                                      signature: bytes):
        """Create authentication response"""
        class AuthenticationResponse:
            def __init__(self, cred_id, cred_raw_id, client_data, auth_data, sig):
                self.id = cred_id
                self.raw_id = cred_raw_id
                self.response = AssertionResponse(client_data, auth_data, sig)
                self.client_extension_results = {}
                self.type = WEBAUTHN_CREDENTIAL_TYPE
                
        class AssertionResponse:
            def __init__(self, client_data, auth_data, sig):
                self.client_data = client_data
                self.authenticator_data = auth_data
                self.signature = sig
                self.user_handle = None
        
        return AuthenticationResponse(
            credential_id_b64,
            credential_id_bytes,
            client_data_json,
            authenticator_data,
            signature
        ) 


#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2025 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import logging
from typing import Dict, Any

from fido2.client import DefaultClientDataCollector
from fido2.webauthn import PublicKeyCredentialCreationOptions, PublicKeyCredentialRequestOptions

from .. import api, utils, rest_api
from ..proto import APIRequest_pb2
from .platforms.detector import BiometricDetector
from .utils.constants import (
    STATUS_SUCCESS, STATUS_BAD_REQUEST, STATUS_ERROR,
    API_ENDPOINTS, API_RESPONSE_MESSAGES
)

WEBAUTHN_ORIGIN_SCHEME = 'https'

class BiometricClient:
    """Main client for biometric authentication operations"""

    def __init__(self):
        self.detector = BiometricDetector()
        self.platform_handler = None
        self._initialize_platform_handler()

    def _initialize_platform_handler(self):
        """Initialize platform-specific handler"""
        try:
            self.platform_handler = self.detector.get_platform_handler()
        except Exception as e:
            logging.warning("Failed to initialize platform handler: %s", str(e))

    def generate_registration_options(self, params, **kwargs) -> Dict[str, Any]:
        """Generate registration options from Keeper API"""
        try:
            rq = APIRequest_pb2.PasskeyRegistrationRequest()
            rq.authenticatorAttachment = APIRequest_pb2.AuthenticatorAttachment.PLATFORM

            rs = api.communicate_rest(
                params,
                rq,
                API_ENDPOINTS['generate_registration'],
                rs_type=APIRequest_pb2.PasskeyRegistrationResponse
            )

            return {
                'challenge_token': rs.challengeToken,
                'creation_options': json.loads(rs.pkCreationOptions)
            }
        except Exception as e:
            raise Exception(str(e))

    def create_credential(self, registration_options: Dict[str, Any]):
        """Create biometric credential"""
        if not self.platform_handler:
            raise Exception("Platform handler not available")

        try:
            creation_options = registration_options['creation_options']

            if isinstance(creation_options.get('challenge'), str):
                creation_options['challenge'] = utils.base64_url_decode(creation_options['challenge'])

            # Handle platform-specific options
            creation_options = self.platform_handler.handle_credential_creation(creation_options)

            # Create WebAuthn client
            options = PublicKeyCredentialCreationOptions.from_dict(creation_options)
            rp_id = options.rp.id or creation_options.get('rp', {}).get('id')
            if not rp_id:
                raise Exception("No RP ID found in API response - server configuration error")
            origin = f'{WEBAUTHN_ORIGIN_SCHEME}://{rp_id}'

            data_collector = DefaultClientDataCollector(origin)
            client = self.platform_handler.create_webauthn_client(data_collector)

            print("Please complete biometric authentication...")
            return self.platform_handler.perform_credential_creation(client, options)

        except Exception as e:
            raise Exception(str(e))

    def verify_registration(self, params, registration_options: Dict[str, Any],
                          credential_response, friendly_name: str):
        """Verify registration with Keeper API"""
        try:
            client_data_bytes = credential_response.response.client_data
            if hasattr(client_data_bytes, 'b64'):
                client_data_b64 = client_data_bytes.b64
            else:
                client_data_b64 = utils.base64_url_encode(client_data_bytes)

            attestation_object = {
                'id': credential_response.id,
                'rawId': utils.base64_url_encode(credential_response.raw_id),
                'response': {
                    'attestationObject': utils.base64_url_encode(credential_response.response.attestation_object),
                    'clientDataJSON': client_data_b64
                },
                'type': 'public-key',
                'clientExtensionResults': credential_response.client_extension_results or {}
            }

            rq = APIRequest_pb2.PasskeyRegistrationFinalization()
            rq.challengeToken = registration_options['challenge_token']
            rq.authenticatorResponse = json.dumps(attestation_object)
            rq.friendlyName = friendly_name

            api.communicate_rest(params, rq, API_ENDPOINTS['verify_registration'])

            if self.platform_handler and hasattr(self.platform_handler, 'storage_handler'):
                storage_handler = getattr(self.platform_handler, 'storage_handler')
                if storage_handler and hasattr(storage_handler, 'store_credential_id'):
                    try:
                        credential_id = credential_response.id
                        success = storage_handler.store_credential_id(params.user, credential_id)
                        if success:
                            logging.debug("Stored credential ID for user: %s", params.user)
                        else:
                            logging.warning("Failed to store credential ID for user: %s", params.user)
                    except Exception as e:
                        logging.warning("Error storing credential ID: %s", str(e))

        except Exception as e:
            raise Exception(str(e))

    def generate_authentication_options(self, params, purpose: str = 'login') -> Dict[str, Any]:
        """Generate authentication options"""
        try:
            rq = APIRequest_pb2.PasskeyAuthenticationRequest()
            rq.authenticatorAttachment = APIRequest_pb2.AuthenticatorAttachment.PLATFORM
            rq.clientVersion = rest_api.CLIENT_VERSION
            rq.username = params.user
            rq.passkeyPurpose = (APIRequest_pb2.PasskeyPurpose.PK_REAUTH
                               if purpose == 'vault' else APIRequest_pb2.PasskeyPurpose.PK_LOGIN)

            if hasattr(params, 'device_token') and params.device_token:
                rq.encryptedDeviceToken = utils.base64_url_decode(params.device_token)

            rs = api.communicate_rest(
                params, rq, API_ENDPOINTS['generate_authentication'],
                rs_type=APIRequest_pb2.PasskeyAuthenticationResponse
            )

            return {
                'challenge_token': rs.challengeToken,
                'request_options': json.loads(rs.pkRequestOptions),
                'login_token': rs.encryptedLoginToken,
                'purpose': purpose
            }
        except Exception as e:
            raise Exception(str(e))

    def perform_authentication(self, auth_options: Dict[str, Any]):
        """Perform biometric authentication"""
        if not self.platform_handler:
            raise Exception("Platform handler not available")

        try:
            request_options = auth_options['request_options']
            pk_options = request_options.get('publicKeyCredentialRequestOptions', request_options)

            if not isinstance(pk_options['challenge'], (bytes, bytearray)):
                pk_options['challenge'] = utils.base64_url_decode(pk_options['challenge'])

            if 'allowCredentials' in pk_options:
                for cred in pk_options['allowCredentials']:
                    if isinstance(cred.get('id'), str):
                        cred['id'] = utils.base64_url_decode(cred['id'])

            pk_options = self.platform_handler.handle_authentication_options(pk_options)

            options = PublicKeyCredentialRequestOptions.from_dict(pk_options)
            rp_id = options.rp_id or pk_options.get('rpId')
            if not rp_id:
                raise Exception("No RP ID found in API response - server configuration error")
            origin = f'{WEBAUTHN_ORIGIN_SCHEME}://{rp_id}'

            data_collector = DefaultClientDataCollector(origin)
            client = self.platform_handler.create_webauthn_client(data_collector)

            return self.platform_handler.perform_authentication(client, options)

        except Exception as e:
            raise Exception(str(e))

    def get_available_credentials(self, params):
        """Get list of available biometric credentials"""
        try:
            rq = APIRequest_pb2.PasskeyListRequest()
            rs = api.communicate_rest(params, rq, API_ENDPOINTS['get_available_keys'], 
                                    rs_type=APIRequest_pb2.PasskeyListResponse)

            return [{
                'id': passkey.userId,
                'name': passkey.friendlyName,
                'created': passkey.createdAtMillis,
                'last_used': passkey.lastUsedMillis,
                'credential_id': passkey.credentialId,
                'aaguid': getattr(passkey, 'AAGUID', None)
            } for passkey in rs.passkeyInfo]

        except Exception as e:
            raise Exception(str(e))

    def disable_passkey(self, params, user_id: int, credential_id: bytes):
        """Disable a passkey using the UpdatePasskeyRequest endpoint"""
        try:
            rq = APIRequest_pb2.UpdatePasskeyRequest()
            rq.userId = user_id
            rq.credentialId = credential_id

            api.communicate_rest(params, rq, API_ENDPOINTS['disable_passkey'])
            
            return {'status': STATUS_SUCCESS, 'message': API_RESPONSE_MESSAGES['passkey_disabled_success']}

        except Exception as e:
            error_msg = str(e).lower()
            if 'bad_request' in error_msg or 'credential id' in error_msg or 'userid' in error_msg:
                return {'status': STATUS_BAD_REQUEST, 'message': API_RESPONSE_MESSAGES['disable_bad_request']}
            elif 'server_error' in error_msg or 'unexpected' in error_msg:
                return {'status': STATUS_ERROR, 'message': API_RESPONSE_MESSAGES['server_exception']}
            else:
                raise Exception(str(e))

    def update_passkey_name(self, params, user_id: int, credential_id: bytes, friendly_name: str):
        """Update the friendly name of a passkey using the UpdatePasskeyRequest endpoint"""
        try:
            rq = APIRequest_pb2.UpdatePasskeyRequest()
            rq.userId = user_id
            rq.credentialId = credential_id
            rq.friendlyName = friendly_name

            api.communicate_rest(params, rq, API_ENDPOINTS['update_passkey_name'])
            
            return {'status': STATUS_SUCCESS, 'message': API_RESPONSE_MESSAGES['passkey_name_updated_success']}

        except Exception as e:
            error_msg = str(e).lower()
            if 'bad_request' in error_msg or 'credential id' in error_msg or 'userid' in error_msg:
                return {'status': STATUS_BAD_REQUEST, 'message': API_RESPONSE_MESSAGES['update_bad_request']}
            elif 'server_error' in error_msg or 'unexpected' in error_msg:
                return {'status': STATUS_ERROR, 'message': API_RESPONSE_MESSAGES['server_exception']}
            else:
                raise Exception(str(e)) 


from google.protobuf import struct_pb2 as _struct_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class Currency(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN: _ClassVar[Currency]
    USD: _ClassVar[Currency]
    GBP: _ClassVar[Currency]
    JPY: _ClassVar[Currency]
    EUR: _ClassVar[Currency]
    AUD: _ClassVar[Currency]
    CAD: _ClassVar[Currency]

class GradientIntegrationStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    NOTCONNECTED: _ClassVar[GradientIntegrationStatus]
    PENDING: _ClassVar[GradientIntegrationStatus]
    CONNECTED: _ClassVar[GradientIntegrationStatus]
    NONE: _ClassVar[GradientIntegrationStatus]

class EventType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = []
    UNKNOWN_TRACKING_EVENT_TYPE: _ClassVar[EventType]
    TRACKING_POPUP_DISPLAYED: _ClassVar[EventType]
    TRACKING_POPUP_ACCEPTED: _ClassVar[EventType]
    TRACKING_POPUP_DISMISSED: _ClassVar[EventType]
    TRACKING_POPUP_PAID: _ClassVar[EventType]
    TRACKING_PUSH_CLICKED: _ClassVar[EventType]
UNKNOWN: Currency
USD: Currency
GBP: Currency
JPY: Currency
EUR: Currency
AUD: Currency
CAD: Currency
NOTCONNECTED: GradientIntegrationStatus
PENDING: GradientIntegrationStatus
CONNECTED: GradientIntegrationStatus
NONE: GradientIntegrationStatus
UNKNOWN_TRACKING_EVENT_TYPE: EventType
TRACKING_POPUP_DISPLAYED: EventType
TRACKING_POPUP_ACCEPTED: EventType
TRACKING_POPUP_DISMISSED: EventType
TRACKING_POPUP_PAID: EventType
TRACKING_PUSH_CLICKED: EventType

class ValidateSessionTokenRequest(_message.Message):
    __slots__ = ["encryptedSessionToken", "returnMcEnterpiseIds", "ip"]
    ENCRYPTEDSESSIONTOKEN_FIELD_NUMBER: _ClassVar[int]
    RETURNMCENTERPISEIDS_FIELD_NUMBER: _ClassVar[int]
    IP_FIELD_NUMBER: _ClassVar[int]
    encryptedSessionToken: bytes
    returnMcEnterpiseIds: bool
    ip: str
    def __init__(self, encryptedSessionToken: _Optional[bytes] = ..., returnMcEnterpiseIds: bool = ..., ip: _Optional[str] = ...) -> None: ...

class ValidateSessionTokenResponse(_message.Message):
    __slots__ = ["username", "userId", "enterpriseUserId", "status", "statusMessage", "mcEnterpriseIds", "hasMSPPermission", "deletedMcEnterpriseIds"]
    class Status(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        VALID: _ClassVar[ValidateSessionTokenResponse.Status]
        NOT_VALID: _ClassVar[ValidateSessionTokenResponse.Status]
        EXPIRED: _ClassVar[ValidateSessionTokenResponse.Status]
        IP_BLOCKED: _ClassVar[ValidateSessionTokenResponse.Status]
        INVALID_CLIENT_VERSION: _ClassVar[ValidateSessionTokenResponse.Status]
    VALID: ValidateSessionTokenResponse.Status
    NOT_VALID: ValidateSessionTokenResponse.Status
    EXPIRED: ValidateSessionTokenResponse.Status
    IP_BLOCKED: ValidateSessionTokenResponse.Status
    INVALID_CLIENT_VERSION: ValidateSessionTokenResponse.Status
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    USERID_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    STATUSMESSAGE_FIELD_NUMBER: _ClassVar[int]
    MCENTERPRISEIDS_FIELD_NUMBER: _ClassVar[int]
    HASMSPPERMISSION_FIELD_NUMBER: _ClassVar[int]
    DELETEDMCENTERPRISEIDS_FIELD_NUMBER: _ClassVar[int]
    username: str
    userId: int
    enterpriseUserId: int
    status: ValidateSessionTokenResponse.Status
    statusMessage: str
    mcEnterpriseIds: _containers.RepeatedScalarFieldContainer[int]
    hasMSPPermission: bool
    deletedMcEnterpriseIds: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, username: _Optional[str] = ..., userId: _Optional[int] = ..., enterpriseUserId: _Optional[int] = ..., status: _Optional[_Union[ValidateSessionTokenResponse.Status, str]] = ..., statusMessage: _Optional[str] = ..., mcEnterpriseIds: _Optional[_Iterable[int]] = ..., hasMSPPermission: bool = ..., deletedMcEnterpriseIds: _Optional[_Iterable[int]] = ...) -> None: ...

class SubscriptionStatusRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SubscriptionStatusResponse(_message.Message):
    __slots__ = ["autoRenewal", "currentPaymentMethod", "checkoutLink", "licenseCreateDate", "isDistributor", "isLegacyMsp", "licenseStats", "gradientStatus", "hideTrialBanner", "gradientLastSyncDate", "gradientNextSyncDate", "isGradientMappingPending"]
    AUTORENEWAL_FIELD_NUMBER: _ClassVar[int]
    CURRENTPAYMENTMETHOD_FIELD_NUMBER: _ClassVar[int]
    CHECKOUTLINK_FIELD_NUMBER: _ClassVar[int]
    LICENSECREATEDATE_FIELD_NUMBER: _ClassVar[int]
    ISDISTRIBUTOR_FIELD_NUMBER: _ClassVar[int]
    ISLEGACYMSP_FIELD_NUMBER: _ClassVar[int]
    LICENSESTATS_FIELD_NUMBER: _ClassVar[int]
    GRADIENTSTATUS_FIELD_NUMBER: _ClassVar[int]
    HIDETRIALBANNER_FIELD_NUMBER: _ClassVar[int]
    GRADIENTLASTSYNCDATE_FIELD_NUMBER: _ClassVar[int]
    GRADIENTNEXTSYNCDATE_FIELD_NUMBER: _ClassVar[int]
    ISGRADIENTMAPPINGPENDING_FIELD_NUMBER: _ClassVar[int]
    autoRenewal: AutoRenewal
    currentPaymentMethod: PaymentMethod
    checkoutLink: str
    licenseCreateDate: int
    isDistributor: bool
    isLegacyMsp: bool
    licenseStats: _containers.RepeatedCompositeFieldContainer[LicenseStats]
    gradientStatus: GradientIntegrationStatus
    hideTrialBanner: bool
    gradientLastSyncDate: str
    gradientNextSyncDate: str
    isGradientMappingPending: bool
    def __init__(self, autoRenewal: _Optional[_Union[AutoRenewal, _Mapping]] = ..., currentPaymentMethod: _Optional[_Union[PaymentMethod, _Mapping]] = ..., checkoutLink: _Optional[str] = ..., licenseCreateDate: _Optional[int] = ..., isDistributor: bool = ..., isLegacyMsp: bool = ..., licenseStats: _Optional[_Iterable[_Union[LicenseStats, _Mapping]]] = ..., gradientStatus: _Optional[_Union[GradientIntegrationStatus, str]] = ..., hideTrialBanner: bool = ..., gradientLastSyncDate: _Optional[str] = ..., gradientNextSyncDate: _Optional[str] = ..., isGradientMappingPending: bool = ...) -> None: ...

class LicenseStats(_message.Message):
    __slots__ = ["type", "available", "used"]
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        LICENSE_STAT_UNKNOWN: _ClassVar[LicenseStats.Type]
        MSP_BASE: _ClassVar[LicenseStats.Type]
        MC_BUSINESS: _ClassVar[LicenseStats.Type]
        MC_BUSINESS_PLUS: _ClassVar[LicenseStats.Type]
        MC_ENTERPRISE: _ClassVar[LicenseStats.Type]
        MC_ENTERPRISE_PLUS: _ClassVar[LicenseStats.Type]
    LICENSE_STAT_UNKNOWN: LicenseStats.Type
    MSP_BASE: LicenseStats.Type
    MC_BUSINESS: LicenseStats.Type
    MC_BUSINESS_PLUS: LicenseStats.Type
    MC_ENTERPRISE: LicenseStats.Type
    MC_ENTERPRISE_PLUS: LicenseStats.Type
    TYPE_FIELD_NUMBER: _ClassVar[int]
    AVAILABLE_FIELD_NUMBER: _ClassVar[int]
    USED_FIELD_NUMBER: _ClassVar[int]
    type: LicenseStats.Type
    available: int
    used: int
    def __init__(self, type: _Optional[_Union[LicenseStats.Type, str]] = ..., available: _Optional[int] = ..., used: _Optional[int] = ...) -> None: ...

class AutoRenewal(_message.Message):
    __slots__ = ["nextOn", "daysLeft", "isTrial"]
    NEXTON_FIELD_NUMBER: _ClassVar[int]
    DAYSLEFT_FIELD_NUMBER: _ClassVar[int]
    ISTRIAL_FIELD_NUMBER: _ClassVar[int]
    nextOn: int
    daysLeft: int
    isTrial: bool
    def __init__(self, nextOn: _Optional[int] = ..., daysLeft: _Optional[int] = ..., isTrial: bool = ...) -> None: ...

class PaymentMethod(_message.Message):
    __slots__ = ["type", "card", "sepa", "paypal", "failedBilling", "vendor", "purchaseOrder"]
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        CARD: _ClassVar[PaymentMethod.Type]
        SEPA: _ClassVar[PaymentMethod.Type]
        PAYPAL: _ClassVar[PaymentMethod.Type]
        NONE: _ClassVar[PaymentMethod.Type]
        VENDOR: _ClassVar[PaymentMethod.Type]
        PURCHASEORDER: _ClassVar[PaymentMethod.Type]
    CARD: PaymentMethod.Type
    SEPA: PaymentMethod.Type
    PAYPAL: PaymentMethod.Type
    NONE: PaymentMethod.Type
    VENDOR: PaymentMethod.Type
    PURCHASEORDER: PaymentMethod.Type
    class Card(_message.Message):
        __slots__ = ["last4", "brand"]
        LAST4_FIELD_NUMBER: _ClassVar[int]
        BRAND_FIELD_NUMBER: _ClassVar[int]
        last4: str
        brand: str
        def __init__(self, last4: _Optional[str] = ..., brand: _Optional[str] = ...) -> None: ...
    class Sepa(_message.Message):
        __slots__ = ["last4", "country"]
        LAST4_FIELD_NUMBER: _ClassVar[int]
        COUNTRY_FIELD_NUMBER: _ClassVar[int]
        last4: str
        country: str
        def __init__(self, last4: _Optional[str] = ..., country: _Optional[str] = ...) -> None: ...
    class Paypal(_message.Message):
        __slots__ = []
        def __init__(self) -> None: ...
    class Vendor(_message.Message):
        __slots__ = ["name"]
        NAME_FIELD_NUMBER: _ClassVar[int]
        name: str
        def __init__(self, name: _Optional[str] = ...) -> None: ...
    class PurchaseOrder(_message.Message):
        __slots__ = ["name"]
        NAME_FIELD_NUMBER: _ClassVar[int]
        name: str
        def __init__(self, name: _Optional[str] = ...) -> None: ...
    TYPE_FIELD_NUMBER: _ClassVar[int]
    CARD_FIELD_NUMBER: _ClassVar[int]
    SEPA_FIELD_NUMBER: _ClassVar[int]
    PAYPAL_FIELD_NUMBER: _ClassVar[int]
    FAILEDBILLING_FIELD_NUMBER: _ClassVar[int]
    VENDOR_FIELD_NUMBER: _ClassVar[int]
    PURCHASEORDER_FIELD_NUMBER: _ClassVar[int]
    type: PaymentMethod.Type
    card: PaymentMethod.Card
    sepa: PaymentMethod.Sepa
    paypal: PaymentMethod.Paypal
    failedBilling: bool
    vendor: PaymentMethod.Vendor
    purchaseOrder: PaymentMethod.PurchaseOrder
    def __init__(self, type: _Optional[_Union[PaymentMethod.Type, str]] = ..., card: _Optional[_Union[PaymentMethod.Card, _Mapping]] = ..., sepa: _Optional[_Union[PaymentMethod.Sepa, _Mapping]] = ..., paypal: _Optional[_Union[PaymentMethod.Paypal, _Mapping]] = ..., failedBilling: bool = ..., vendor: _Optional[_Union[PaymentMethod.Vendor, _Mapping]] = ..., purchaseOrder: _Optional[_Union[PaymentMethod.PurchaseOrder, _Mapping]] = ...) -> None: ...

class SubscriptionMspPricingRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SubscriptionMspPricingResponse(_message.Message):
    __slots__ = ["addons", "filePlans"]
    ADDONS_FIELD_NUMBER: _ClassVar[int]
    FILEPLANS_FIELD_NUMBER: _ClassVar[int]
    addons: _containers.RepeatedCompositeFieldContainer[Addon]
    filePlans: _containers.RepeatedCompositeFieldContainer[FilePlan]
    def __init__(self, addons: _Optional[_Iterable[_Union[Addon, _Mapping]]] = ..., filePlans: _Optional[_Iterable[_Union[FilePlan, _Mapping]]] = ...) -> None: ...

class SubscriptionMcPricingRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class SubscriptionMcPricingResponse(_message.Message):
    __slots__ = ["basePlans", "addons", "filePlans"]
    BASEPLANS_FIELD_NUMBER: _ClassVar[int]
    ADDONS_FIELD_NUMBER: _ClassVar[int]
    FILEPLANS_FIELD_NUMBER: _ClassVar[int]
    basePlans: _containers.RepeatedCompositeFieldContainer[BasePlan]
    addons: _containers.RepeatedCompositeFieldContainer[Addon]
    filePlans: _containers.RepeatedCompositeFieldContainer[FilePlan]
    def __init__(self, basePlans: _Optional[_Iterable[_Union[BasePlan, _Mapping]]] = ..., addons: _Optional[_Iterable[_Union[Addon, _Mapping]]] = ..., filePlans: _Optional[_Iterable[_Union[FilePlan, _Mapping]]] = ...) -> None: ...

class BasePlan(_message.Message):
    __slots__ = ["id", "cost"]
    ID_FIELD_NUMBER: _ClassVar[int]
    COST_FIELD_NUMBER: _ClassVar[int]
    id: int
    cost: Cost
    def __init__(self, id: _Optional[int] = ..., cost: _Optional[_Union[Cost, _Mapping]] = ...) -> None: ...

class Addon(_message.Message):
    __slots__ = ["id", "cost", "amountConsumed"]
    ID_FIELD_NUMBER: _ClassVar[int]
    COST_FIELD_NUMBER: _ClassVar[int]
    AMOUNTCONSUMED_FIELD_NUMBER: _ClassVar[int]
    id: int
    cost: Cost
    amountConsumed: int
    def __init__(self, id: _Optional[int] = ..., cost: _Optional[_Union[Cost, _Mapping]] = ..., amountConsumed: _Optional[int] = ...) -> None: ...

class FilePlan(_message.Message):
    __slots__ = ["id", "cost"]
    ID_FIELD_NUMBER: _ClassVar[int]
    COST_FIELD_NUMBER: _ClassVar[int]
    id: int
    cost: Cost
    def __init__(self, id: _Optional[int] = ..., cost: _Optional[_Union[Cost, _Mapping]] = ...) -> None: ...

class Cost(_message.Message):
    __slots__ = ["amount", "amountPer", "currency"]
    class AmountPer(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        UNKNOWN: _ClassVar[Cost.AmountPer]
        MONTH: _ClassVar[Cost.AmountPer]
        USER_MONTH: _ClassVar[Cost.AmountPer]
        USER_CONSUMED_MONTH: _ClassVar[Cost.AmountPer]
        ENDPOINT_MONTH: _ClassVar[Cost.AmountPer]
    UNKNOWN: Cost.AmountPer
    MONTH: Cost.AmountPer
    USER_MONTH: Cost.AmountPer
    USER_CONSUMED_MONTH: Cost.AmountPer
    ENDPOINT_MONTH: Cost.AmountPer
    AMOUNT_FIELD_NUMBER: _ClassVar[int]
    AMOUNTPER_FIELD_NUMBER: _ClassVar[int]
    CURRENCY_FIELD_NUMBER: _ClassVar[int]
    amount: float
    amountPer: Cost.AmountPer
    currency: Currency
    def __init__(self, amount: _Optional[float] = ..., amountPer: _Optional[_Union[Cost.AmountPer, str]] = ..., currency: _Optional[_Union[Currency, str]] = ...) -> None: ...

class InvoiceSearchRequest(_message.Message):
    __slots__ = ["size", "startingAfterId", "allInvoicesUnfiltered"]
    SIZE_FIELD_NUMBER: _ClassVar[int]
    STARTINGAFTERID_FIELD_NUMBER: _ClassVar[int]
    ALLINVOICESUNFILTERED_FIELD_NUMBER: _ClassVar[int]
    size: int
    startingAfterId: int
    allInvoicesUnfiltered: bool
    def __init__(self, size: _Optional[int] = ..., startingAfterId: _Optional[int] = ..., allInvoicesUnfiltered: bool = ...) -> None: ...

class InvoiceSearchResponse(_message.Message):
    __slots__ = ["invoices"]
    INVOICES_FIELD_NUMBER: _ClassVar[int]
    invoices: _containers.RepeatedCompositeFieldContainer[Invoice]
    def __init__(self, invoices: _Optional[_Iterable[_Union[Invoice, _Mapping]]] = ...) -> None: ...

class Invoice(_message.Message):
    __slots__ = ["id", "invoiceNumber", "invoiceDate", "licenseCount", "totalCost", "invoiceType"]
    class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = []
        UNKNOWN: _ClassVar[Invoice.Type]
        NEW: _ClassVar[Invoice.Type]
        RENEWAL: _ClassVar[Invoice.Type]
        UPGRADE: _ClassVar[Invoice.Type]
        RESTORE: _ClassVar[Invoice.Type]
        ASSOCIATION: _ClassVar[Invoice.Type]
        OVERAGE: _ClassVar[Invoice.Type]
    UNKNOWN: Invoice.Type
    NEW: Invoice.Type
    RENEWAL: Invoice.Type
    UPGRADE: Invoice.Type
    RESTORE: Invoice.Type
    ASSOCIATION: Invoice.Type
    OVERAGE: Invoice.Type
    class Cost(_message.Message):
        __slots__ = ["amount", "currency"]
        AMOUNT_FIELD_NUMBER: _ClassVar[int]
        CURRENCY_FIELD_NUMBER: _ClassVar[int]
        amount: float
        currency: Currency
        def __init__(self, amount: _Optional[float] = ..., currency: _Optional[_Union[Currency, str]] = ...) -> None: ...
    ID_FIELD_NUMBER: _ClassVar[int]
    INVOICENUMBER_FIELD_NUMBER: _ClassVar[int]
    INVOICEDATE_FIELD_NUMBER: _ClassVar[int]
    LICENSECOUNT_FIELD_NUMBER: _ClassVar[int]
    TOTALCOST_FIELD_NUMBER: _ClassVar[int]
    INVOICETYPE_FIELD_NUMBER: _ClassVar[int]
    id: int
    invoiceNumber: str
    invoiceDate: int
    licenseCount: int
    totalCost: Invoice.Cost
    invoiceType: Invoice.Type
    def __init__(self, id: _Optional[int] = ..., invoiceNumber: _Optional[str] = ..., invoiceDate: _Optional[int] = ..., licenseCount: _Optional[int] = ..., totalCost: _Optional[_Union[Invoice.Cost, _Mapping]] = ..., invoiceType: _Optional[_Union[Invoice.Type, str]] = ...) -> None: ...

class VaultInvoicesListRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class VaultInvoicesListResponse(_message.Message):
    __slots__ = ["invoices"]
    INVOICES_FIELD_NUMBER: _ClassVar[int]
    invoices: _containers.RepeatedCompositeFieldContainer[VaultInvoice]
    def __init__(self, invoices: _Optional[_Iterable[_Union[VaultInvoice, _Mapping]]] = ...) -> None: ...

class VaultInvoice(_message.Message):
    __slots__ = ["id", "invoiceNumber", "dateCreated", "total", "purchaseType"]
    ID_FIELD_NUMBER: _ClassVar[int]
    INVOICENUMBER_FIELD_NUMBER: _ClassVar[int]
    DATECREATED_FIELD_NUMBER: _ClassVar[int]
    TOTAL_FIELD_NUMBER: _ClassVar[int]
    PURCHASETYPE_FIELD_NUMBER: _ClassVar[int]
    id: int
    invoiceNumber: str
    dateCreated: int
    total: Invoice.Cost
    purchaseType: Invoice.Type
    def __init__(self, id: _Optional[int] = ..., invoiceNumber: _Optional[str] = ..., dateCreated: _Optional[int] = ..., total: _Optional[_Union[Invoice.Cost, _Mapping]] = ..., purchaseType: _Optional[_Union[Invoice.Type, str]] = ...) -> None: ...

class InvoiceDownloadRequest(_message.Message):
    __slots__ = ["invoiceNumber"]
    INVOICENUMBER_FIELD_NUMBER: _ClassVar[int]
    invoiceNumber: str
    def __init__(self, invoiceNumber: _Optional[str] = ...) -> None: ...

class InvoiceDownloadResponse(_message.Message):
    __slots__ = ["link", "fileName"]
    LINK_FIELD_NUMBER: _ClassVar[int]
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    link: str
    fileName: str
    def __init__(self, link: _Optional[str] = ..., fileName: _Optional[str] = ...) -> None: ...

class VaultInvoiceDownloadLinkRequest(_message.Message):
    __slots__ = ["invoiceNumber"]
    INVOICENUMBER_FIELD_NUMBER: _ClassVar[int]
    invoiceNumber: str
    def __init__(self, invoiceNumber: _Optional[str] = ...) -> None: ...

class VaultInvoiceDownloadLinkResponse(_message.Message):
    __slots__ = ["link", "fileName"]
    LINK_FIELD_NUMBER: _ClassVar[int]
    FILENAME_FIELD_NUMBER: _ClassVar[int]
    link: str
    fileName: str
    def __init__(self, link: _Optional[str] = ..., fileName: _Optional[str] = ...) -> None: ...

class ReportingDailySnapshotRequest(_message.Message):
    __slots__ = ["month", "year"]
    MONTH_FIELD_NUMBER: _ClassVar[int]
    YEAR_FIELD_NUMBER: _ClassVar[int]
    month: int
    year: int
    def __init__(self, month: _Optional[int] = ..., year: _Optional[int] = ...) -> None: ...

class ReportingDailySnapshotResponse(_message.Message):
    __slots__ = ["records", "mcEnterprises"]
    RECORDS_FIELD_NUMBER: _ClassVar[int]
    MCENTERPRISES_FIELD_NUMBER: _ClassVar[int]
    records: _containers.RepeatedCompositeFieldContainer[SnapshotRecord]
    mcEnterprises: _containers.RepeatedCompositeFieldContainer[SnapshotMcEnterprise]
    def __init__(self, records: _Optional[_Iterable[_Union[SnapshotRecord, _Mapping]]] = ..., mcEnterprises: _Optional[_Iterable[_Union[SnapshotMcEnterprise, _Mapping]]] = ...) -> None: ...

class SnapshotRecord(_message.Message):
    __slots__ = ["date", "mcEnterpriseId", "maxLicenseCount", "maxFilePlanTypeId", "maxBasePlanId", "addons"]
    class Addon(_message.Message):
        __slots__ = ["maxAddonId", "units"]
        MAXADDONID_FIELD_NUMBER: _ClassVar[int]
        UNITS_FIELD_NUMBER: _ClassVar[int]
        maxAddonId: int
        units: int
        def __init__(self, maxAddonId: _Optional[int] = ..., units: _Optional[int] = ...) -> None: ...
    DATE_FIELD_NUMBER: _ClassVar[int]
    MCENTERPRISEID_FIELD_NUMBER: _ClassVar[int]
    MAXLICENSECOUNT_FIELD_NUMBER: _ClassVar[int]
    MAXFILEPLANTYPEID_FIELD_NUMBER: _ClassVar[int]
    MAXBASEPLANID_FIELD_NUMBER: _ClassVar[int]
    ADDONS_FIELD_NUMBER: _ClassVar[int]
    date: int
    mcEnterpriseId: int
    maxLicenseCount: int
    maxFilePlanTypeId: int
    maxBasePlanId: int
    addons: _containers.RepeatedCompositeFieldContainer[SnapshotRecord.Addon]
    def __init__(self, date: _Optional[int] = ..., mcEnterpriseId: _Optional[int] = ..., maxLicenseCount: _Optional[int] = ..., maxFilePlanTypeId: _Optional[int] = ..., maxBasePlanId: _Optional[int] = ..., addons: _Optional[_Iterable[_Union[SnapshotRecord.Addon, _Mapping]]] = ...) -> None: ...

class SnapshotMcEnterprise(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    id: int
    name: str
    def __init__(self, id: _Optional[int] = ..., name: _Optional[str] = ...) -> None: ...

class MappingAddonsRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class MappingAddonsResponse(_message.Message):
    __slots__ = ["addons", "filePlans"]
    ADDONS_FIELD_NUMBER: _ClassVar[int]
    FILEPLANS_FIELD_NUMBER: _ClassVar[int]
    addons: _containers.RepeatedCompositeFieldContainer[MappingItem]
    filePlans: _containers.RepeatedCompositeFieldContainer[MappingItem]
    def __init__(self, addons: _Optional[_Iterable[_Union[MappingItem, _Mapping]]] = ..., filePlans: _Optional[_Iterable[_Union[MappingItem, _Mapping]]] = ...) -> None: ...

class MappingItem(_message.Message):
    __slots__ = ["id", "name"]
    ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    id: int
    name: str
    def __init__(self, id: _Optional[int] = ..., name: _Optional[str] = ...) -> None: ...

class GradientValidateKeyRequest(_message.Message):
    __slots__ = ["gradientKey"]
    GRADIENTKEY_FIELD_NUMBER: _ClassVar[int]
    gradientKey: str
    def __init__(self, gradientKey: _Optional[str] = ...) -> None: ...

class GradientValidateKeyResponse(_message.Message):
    __slots__ = ["success", "message"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class GradientSaveRequest(_message.Message):
    __slots__ = ["gradientKey", "enterpriseUserId"]
    GRADIENTKEY_FIELD_NUMBER: _ClassVar[int]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    gradientKey: str
    enterpriseUserId: int
    def __init__(self, gradientKey: _Optional[str] = ..., enterpriseUserId: _Optional[int] = ...) -> None: ...

class GradientSaveResponse(_message.Message):
    __slots__ = ["success", "status", "message"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    status: GradientIntegrationStatus
    message: str
    def __init__(self, success: bool = ..., status: _Optional[_Union[GradientIntegrationStatus, str]] = ..., message: _Optional[str] = ...) -> None: ...

class GradientRemoveRequest(_message.Message):
    __slots__ = ["enterpriseUserId"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    def __init__(self, enterpriseUserId: _Optional[int] = ...) -> None: ...

class GradientRemoveResponse(_message.Message):
    __slots__ = ["success", "message"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    message: str
    def __init__(self, success: bool = ..., message: _Optional[str] = ...) -> None: ...

class GradientSyncRequest(_message.Message):
    __slots__ = ["enterpriseUserId"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    def __init__(self, enterpriseUserId: _Optional[int] = ...) -> None: ...

class GradientSyncResponse(_message.Message):
    __slots__ = ["success", "status", "message"]
    SUCCESS_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    success: bool
    status: GradientIntegrationStatus
    message: str
    def __init__(self, success: bool = ..., status: _Optional[_Union[GradientIntegrationStatus, str]] = ..., message: _Optional[str] = ...) -> None: ...

class NetPromoterScoreSurveySubmissionRequest(_message.Message):
    __slots__ = ["survey_score", "notes"]
    SURVEY_SCORE_FIELD_NUMBER: _ClassVar[int]
    NOTES_FIELD_NUMBER: _ClassVar[int]
    survey_score: int
    notes: str
    def __init__(self, survey_score: _Optional[int] = ..., notes: _Optional[str] = ...) -> None: ...

class NetPromoterScoreSurveySubmissionResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class NetPromoterScorePopupScheduleRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class NetPromoterScorePopupScheduleResponse(_message.Message):
    __slots__ = ["show_popup"]
    SHOW_POPUP_FIELD_NUMBER: _ClassVar[int]
    show_popup: bool
    def __init__(self, show_popup: bool = ...) -> None: ...

class NetPromoterScorePopupDismissalRequest(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class NetPromoterScorePopupDismissalResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...

class KCMLicenseRequest(_message.Message):
    __slots__ = ["enterpriseUserId"]
    ENTERPRISEUSERID_FIELD_NUMBER: _ClassVar[int]
    enterpriseUserId: int
    def __init__(self, enterpriseUserId: _Optional[int] = ...) -> None: ...

class KCMLicenseResponse(_message.Message):
    __slots__ = ["message"]
    MESSAGE_FIELD_NUMBER: _ClassVar[int]
    message: str
    def __init__(self, message: _Optional[str] = ...) -> None: ...

class EventRequest(_message.Message):
    __slots__ = ["eventType", "eventValue", "eventTime", "attributes"]
    EVENTTYPE_FIELD_NUMBER: _ClassVar[int]
    EVENTVALUE_FIELD_NUMBER: _ClassVar[int]
    EVENTTIME_FIELD_NUMBER: _ClassVar[int]
    ATTRIBUTES_FIELD_NUMBER: _ClassVar[int]
    eventType: EventType
    eventValue: str
    eventTime: int
    attributes: _struct_pb2.Struct
    def __init__(self, eventType: _Optional[_Union[EventType, str]] = ..., eventValue: _Optional[str] = ..., eventTime: _Optional[int] = ..., attributes: _Optional[_Union[_struct_pb2.Struct, _Mapping]] = ...) -> None: ...

class EventsRequest(_message.Message):
    __slots__ = ["event"]
    EVENT_FIELD_NUMBER: _ClassVar[int]
    event: _containers.RepeatedCompositeFieldContainer[EventRequest]
    def __init__(self, event: _Optional[_Iterable[_Union[EventRequest, _Mapping]]] = ...) -> None: ...

class CustomerCaptureRequest(_message.Message):
    __slots__ = ["pageUrl", "tree", "hash", "image", "pageLoadTime", "keyId", "test"]
    PAGEURL_FIELD_NUMBER: _ClassVar[int]
    TREE_FIELD_NUMBER: _ClassVar[int]
    HASH_FIELD_NUMBER: _ClassVar[int]
    IMAGE_FIELD_NUMBER: _ClassVar[int]
    PAGELOADTIME_FIELD_NUMBER: _ClassVar[int]
    KEYID_FIELD_NUMBER: _ClassVar[int]
    TEST_FIELD_NUMBER: _ClassVar[int]
    pageUrl: str
    tree: str
    hash: str
    image: str
    pageLoadTime: str
    keyId: str
    test: bool
    def __init__(self, pageUrl: _Optional[str] = ..., tree: _Optional[str] = ..., hash: _Optional[str] = ..., image: _Optional[str] = ..., pageLoadTime: _Optional[str] = ..., keyId: _Optional[str] = ..., test: bool = ...) -> None: ...

class CustomerCaptureResponse(_message.Message):
    __slots__ = []
    def __init__(self) -> None: ...



import enum
import sys
from typing import List, Tuple, Optional
from datetime import timedelta
from urllib.parse import urlparse

ENTERPRISE_FILE_PLANS = [
    (-1, 'NO_STORAGE', 'NO STORAGE'),
    (0, 'STORAGE_TRIAL', 'TRIAL'),
    (1, 'STORAGE_1GB', '1GB'),
    (2, 'STORAGE_10GB', '10GB'),
    (3, 'STORAGE_50GB', '50GB'),
    (4, 'STORAGE_100GB', '100GB'),
    (5, 'STORAGE_250GB', '250GB'),
    (6, 'STORAGE_500GB', '500GB'),
    (7, 'STORAGE_1000GB', '1TB'),
    (8, 'STORAGE_10000GB', '10TB'),
]

MSP_FILE_PLANS = [
    (4, 'STORAGE_100GB', '100GB'),
    (7, 'STORAGE_1000GB', '1TB'),
    (8, 'STORAGE_10000GB', '10TB'),
]

MSP_PLANS = [
    (1, 'business', 'Business', 4),
    (2, 'businessPlus', 'Business Plus', 7),
    (3, 'businessStarter', 'Business Starter', 0),
    (10, 'enterprise', 'Enterprise', 4),
    (11, 'enterprisePlus', 'Enterprise Plus', 7),
]

MSP_ADDONS = [
    ('chat', 'KeeperChat', False, 'Chat'),
    ('enterprise_audit_and_reporting', 'Advanced Reporting & Alerts Module', False, 'ARAM'),
    ('professional_services_silver_add_on', 'Professional Services & Support Silver Plan', False, 'Silver Support'),
    ('gold_professional_services_add_on', 'Professional Services & Support Gold Plan', False, 'Gold Support'),
    ('platinum_professional_services_add_on', 'Professional Services & Support Platinum Plan', False, 'Platinum Support'),
    ('msp_service_and_support', 'MSP Dedicated Service & Support', False, 'MSP Support'),
    ('consumer_breach_watch', 'Consumer BreachWatch', False, 'Consumer BW'),
    ('enterprise_breach_watch', 'Enterprise BreachWatch', False, 'Enterprise BW'),
    ('compliance_report', 'Compliance Reporting', False, 'Compliance'),
    ('secrets_manager', 'Keeper Secrets Manager (KSM)', False, 'KSM'),
    ('connection_manager', 'Keeper Connection Manager (KCM)', True, 'KCM'),
    ('password_rotation', 'Password Rotation', False, 'Rotation'),
    ('remote_browser_isolation', 'Remote Browser Isolation', False, 'Browser Isolation'),
    ('privileged_access_manager', 'Privileged Access Manager (PAM)', True, 'PAM'),
]


class PrivilegeScope(enum.IntEnum):
    All = 1,
    MSP = 2,
    Hidden = 3,


# Managed Role privileges
ROLE_PRIVILEGES = [
    ('Manage Nodes', 'MANAGE_NODES', PrivilegeScope.All),
    ('Manage Users', 'MANAGE_USER', PrivilegeScope.All),
    ('Manage Licences', 'MANAGE_LICENCES', PrivilegeScope.Hidden),
    ('Manage Roles', 'MANAGE_ROLES', PrivilegeScope.All),
    ('Manage Teams', 'MANAGE_TEAMS', PrivilegeScope.All),
    ('Run Security Reports', 'RUN_REPORTS', PrivilegeScope.All),
    ('Manage Bridge/SSO', 'MANAGE_BRIDGE', PrivilegeScope.All),
    ('Perform Device Approvals', 'APPROVE_DEVICE', PrivilegeScope.All),
    ('Manage Record Types in Vault', 'MANAGE_RECORD_TYPES', PrivilegeScope.All),
    ('Run Compliance Reports', 'RUN_COMPLIANCE_REPORTS', PrivilegeScope.All),
    ('Manage Companies', 'MANAGE_COMPANIES', PrivilegeScope.MSP),
    ('Transfer Account', 'TRANSFER_ACCOUNT', PrivilegeScope.All),
    ('Sharing Administrator', 'SHARING_ADMINISTRATOR', PrivilegeScope.All),
]

# Timeout constants
# Set to default value by using timedelta of 0
TIMEOUT_DEFAULT = timedelta(0)
TIMEOUT_MIN = timedelta(minutes=1)
TIMEOUT_DEFAULT_UNIT = 'minutes'
TIMEOUT_ALLOWED_UNITS = (
    ('years', 'y'),
    ('months', 'mo'),
    ('days', 'd'),
    ('hours', 'h'),
    ('minutes', 'mi')
)

EMAIL_PATTERN = r"(^[a-zA-Z0-9_.+-]+@[a-zA-Z0-9-]+\.[a-zA-Z0-9-.]+$)"

# Enforcement constants

_ENFORCEMENT_GROUPS = [
    "LOGIN_SETTINGS",
    "TWO_FACTOR_AUTHENTICATION",
    "PLATFORM_RESTRICTION",
    "VAULT_FEATURES",
    "RECORD_TYPES",
    "SHARING_AND_UPLOADING",
    "CREATING_AND_SHARING",
    "KEEPER_FILL",
    "ACCOUNT_SETTINGS",
    "ALLOW_IP_LIST",
]

_ENFORCEMENTS = [
    ("MASTER_PASSWORD_MINIMUM_LENGTH", 10, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_MINIMUM_SPECIAL", 11, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_MINIMUM_UPPER", 12, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_MINIMUM_LOWER", 13, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_MINIMUM_DIGITS", 14, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_RESTRICT_DAYS_BEFORE_REUSE", 16, "LONG", "LOGIN_SETTINGS"),
    ("REQUIRE_TWO_FACTOR", 20, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("MASTER_PASSWORD_MAXIMUM_DAYS_BEFORE_CHANGE", 22, "LONG", "LOGIN_SETTINGS"),
    ("MASTER_PASSWORD_EXPIRED_AS_OF", 23, "LONG", "LOGIN_SETTINGS"),
    ("MINIMUM_PBKDF2_ITERATIONS", 55, "LONG", "ACCOUNT_SETTINGS"),
    ("MAX_SESSION_LOGIN_TIME", 24, "LONG", "ACCOUNT_SETTINGS"),
    ("RESTRICT_PERSISTENT_LOGIN", 25, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("STAY_LOGGED_IN_DEFAULT", 26, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_SHARING_ALL_OUTGOING", 30, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_SHARING_ENTERPRISE_OUTGOING", 31, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_EXPORT", 32, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_FILE_UPLOAD", 33, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("REQUIRE_ACCOUNT_SHARE", 34, "ACCOUNT_SHARE", "SHARING_AND_UPLOADING"),
    ("RESTRICT_SHARING_ALL_INCOMING", 36, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_SHARING_ENTERPRISE_INCOMING", 37, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_IP_ADDRESSES", 40, "IP_WHITELIST", "ALLOW_IP_LIST"),
    ("REQUIRE_DEVICE_APPROVAL", 41, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("REQUIRE_ACCOUNT_RECOVERY_APPROVAL", 42, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("RESTRICT_VAULT_IP_ADDRESSES", 43, "IP_WHITELIST", "ALLOW_IP_LIST"),
    ("TIP_ZONE_RESTRICT_ALLOWED_IP_RANGES", 44, "IP_WHITELIST", "ALLOW_IP_LIST"),
    ("AUTOMATIC_BACKUP_EVERY_X_DAYS", 45, "LONG", "ACCOUNT_SETTINGS"),
    ("RESTRICT_OFFLINE_ACCESS", 46, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("SEND_INVITE_AT_REGISTRATION", 47, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("RESTRICT_EMAIL_CHANGE", 48, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("RESTRICT_IOS_FINGERPRINT", 49, "BOOLEAN", "LOGIN_SETTINGS"),
    ("RESTRICT_MAC_FINGERPRINT", 50, "BOOLEAN", "LOGIN_SETTINGS"),
    ("RESTRICT_ANDROID_FINGERPRINT", 51, "BOOLEAN", "LOGIN_SETTINGS"),
    ("LOGOUT_TIMER_WEB", 52, "LONG", "ACCOUNT_SETTINGS"),
    ("LOGOUT_TIMER_MOBILE", 53, "LONG", "ACCOUNT_SETTINGS"),
    ("LOGOUT_TIMER_DESKTOP", 54, "LONG", "ACCOUNT_SETTINGS"),
    ("RESTRICT_WEB_VAULT_ACCESS", 60, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_EXTENSIONS_ACCESS", 61, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_MOBILE_ACCESS", 62, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_DESKTOP_ACCESS", 63, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_MOBILE_IOS_ACCESS", 64, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_MOBILE_ANDROID_ACCESS", 65, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_MOBILE_WINDOWS_PHONE_ACCESS", 66, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_DESKTOP_WIN_ACCESS", 67, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_DESKTOP_MAC_ACCESS", 68, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_CHAT_DESKTOP_ACCESS", 84, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_CHAT_MOBILE_ACCESS", 85, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_COMMANDER_ACCESS", 88, "BOOLEAN", "PLATFORM_RESTRICTION"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_TEXT", 70, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_GOOGLE", 71, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_DNA", 72, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_DUO", 73, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_RSA", 74, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("TWO_FACTOR_DURATION_WEB", 80, "TWO_FACTOR_DURATION", "TWO_FACTOR_AUTHENTICATION"),
    ("TWO_FACTOR_DURATION_MOBILE", 81, "TWO_FACTOR_DURATION", "TWO_FACTOR_AUTHENTICATION"),
    ("TWO_FACTOR_DURATION_DESKTOP", 82, "TWO_FACTOR_DURATION", "TWO_FACTOR_AUTHENTICATION"),
    ("RESTRICT_WINDOWS_FINGERPRINT", 83, "BOOLEAN", "LOGIN_SETTINGS"),
    ("RESTRICT_TWO_FACTOR_CHANNEL_SECURITY_KEYS", 86, "BOOLEAN", "TWO_FACTOR_AUTHENTICATION"),
    ("TWO_FACTOR_BY_IP", 87, "JSONARRAY"),
    ("RESTRICT_DOMAIN_ACCESS", 90, "STRING", "KEEPER_FILL"),
    ("RESTRICT_DOMAIN_CREATE", 91, "STRING", "KEEPER_FILL"),
    ("RESTRICT_HOVER_LOCKS", 92, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_PROMPT_TO_LOGIN", 93, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_PROMPT_TO_FILL", 94, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_AUTO_SUBMIT", 95, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_PROMPT_TO_SAVE", 96, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_PROMPT_TO_CHANGE", 97, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_AUTO_FILL", 98, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_CREATE_FOLDER", 100, "BOOLEAN", "VAULT_FEATURES"),
    ("RESTRICT_CREATE_FOLDER_TO_ONLY_SHARED_FOLDERS", 101, "BOOLEAN", "CREATING_AND_SHARING"),
    ("RESTRICT_CREATE_IDENTITY_PAYMENT_RECORDS", 102, "BOOLEAN", "VAULT_FEATURES"),
    ("MASK_CUSTOM_FIELDS", 103, "BOOLEAN", "VAULT_FEATURES"),
    ("MASK_NOTES", 104, "BOOLEAN", "VAULT_FEATURES"),
    ("MASK_PASSWORDS_WHILE_EDITING", 105, "BOOLEAN", "VAULT_FEATURES"),
    ("GENERATED_PASSWORD_COMPLEXITY", 106, "PASSWORD_COMPLEXITY", "VAULT_FEATURES"),
    ("GENERATED_SECURITY_QUESTION_COMPLEXITY", 109, "STRING", "VAULT_FEATURES"),
    ("DAYS_BEFORE_DELETED_RECORDS_CLEARED_PERM", 107, "LONG", "VAULT_FEATURES"),
    ("DAYS_BEFORE_DELETED_RECORDS_AUTO_CLEARED", 108, "LONG", "VAULT_FEATURES"),
    ("ALLOW_ALTERNATE_PASSWORDS", 110, "BOOLEAN", "LOGIN_SETTINGS"),
    ("RESTRICT_IMPORT", 111, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_CREATE_RECORD", 112, "BOOLEAN", "CREATING_AND_SHARING"),
    ("RESTRICT_CREATE_RECORD_TO_SHARED_FOLDERS", 113, "BOOLEAN", "CREATING_AND_SHARING"),
    ("RESTRICT_CREATE_SHARED_FOLDER", 114, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_SHARING_RECORD_WITH_ATTACHMENTS", 121, "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_LINK_SHARING", 122, "BOOLEAN", "SHARING_ENFORCEMENTS"),
    ("RESTRICT_SHARING_OUTSIDE_OF_ISOLATED_NODES", 123, "BOOLEAN", "SHARING_ENFORCEMENTS"),
    ("RESTRICT_SHARING_RECORD_TO_SHARED_FOLDERS", 124, "BOOLEAN", "SHARING_ENFORCEMENTS"),
    ("DISABLE_SETUP_TOUR", 140, "BOOLEAN", "VAULT_FEATURES"),
    ("RESTRICT_PERSONAL_LICENSE", 141, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("DISABLE_ONBOARDING", 142, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("DISALLOW_V2_CLIENTS", 143, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("RESTRICT_IP_AUTOAPPROVAL", 144, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("SEND_BREACH_WATCH_EVENTS", 200, "BOOLEAN", "VAULT_FEATURES"),
    ("RESTRICT_BREACH_WATCH", 201, "BOOLEAN", "VAULT_FEATURES"),
    ("RESEND_ENTERPRISE_INVITE_IN_X_DAYS", 202, "LONG", "ACCOUNT_SETTINGS"),
    ("MASTER_PASSWORD_REENTRY", 203, "JSON"),
    ("RESTRICT_ACCOUNT_RECOVERY", 204, "BOOLEAN", "ACCOUNT_SETTINGS"),
    ("KEEPER_FILL_HOVER_LOCKS", 205, "TERNARY_DEN", "KEEPER_FILL"),
    ("KEEPER_FILL_AUTO_FILL", 206, "TERNARY_DEN", "KEEPER_FILL"),
    ("KEEPER_FILL_AUTO_SUBMIT", 207, "TERNARY_DEN", "KEEPER_FILL"),
    ("KEEPER_FILL_MATCH_ON_SUBDOMAIN", 208, "TERNARY_EDN", "KEEPER_FILL"),
    ("RESTRICT_PROMPT_TO_DISABLE", 209, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_HTTP_FILL_WARNING", 210, "BOOLEAN", "KEEPER_FILL"),
    ("RESTRICT_RECORD_TYPES", 211, "RECORD_TYPES", "RECORD_TYPES"),
    ("ALLOW_SECRETS_MANAGER", 212, "BOOLEAN", "VAULT_FEATURES"),
    ("REQUIRE_SELF_DESTRUCT", 213, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("KEEPER_FILL_AUTO_SUGGEST", 214, "TERNARY_DEN", "KEEPER_FILL"),
    ("MAXIMUM_RECORD_SIZE", 215, "LONG", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_PAM_ROTATION", 218, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_PAM_DISCOVERY", 219, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_IMPORT_SHARED_FOLDERS", 220, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("REQUIRE_SECURITY_KEY_PIN", 221, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("DISABLE_CREATE_DUPLICATE", 224, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_PAM_GATEWAY", 225, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_CONFIGURE_ROTATION_SETTINGS", 226, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_ROTATE_CREDENTIALS", 227, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_CONFIGURE_PAM_CLOUD_CONNECTION_SETTINGS", 228, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_LAUNCH_PAM_ON_CLOUD_CONNECTION", 229, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_CONFIGURE_PAM_TUNNELING_SETTINGS", 230, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_LAUNCH_PAM_TUNNELS", 231, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_LAUNCH_RBI", 232, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_CONFIGURE_RBI", 233, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_VIEW_KCM_RECORDINGS", 234, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_TOTP_FIELD", 235, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("ALLOW_VIEW_RBI_RECORDINGS", 236, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_MANAGE_TLA", 238, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_SELF_DESTRUCT_RECORDS", 239, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_PERSONAL_USING_BUSINESS_DOMAINS", 240, "STRING", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_BUSINESS_USING_PERSONAL_USERNAME", 241, "LONG", "ACCOUNT_ENFORCEMENTS"),
    ("WARN_PERSONAL_USING_BUSINESS_DOMAINS", 242, "STRING", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_PERSONAL_USING_BUSINESS_SITES", 243, "STRING", "ACCOUNT_ENFORCEMENTS"),
    ("WARN_PERSONAL_USING_BUSINESS_SITES", 244, "STRING", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_ACCOUNT_SWITCHING", 245, "BOOLEAN", "AUTHENTICATION_ENFORCEMENTS"),
    ("RESTRICT_PASSKEY_LOGIN", 246, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
    ("RESTRICT_CAN_EDIT_EXTERNAL_SHARES", 247, "BOOLEAN", "ACCOUNT_ENFORCEMENTS"),
]

_COMPOUND_ENFORCEMENTS = [
    ("RESTRICT_SHARING_ALL", (30, 36), "BOOLEAN", "SHARING_AND_UPLOADING"),
    ("RESTRICT_SHARING_ENTERPRISE", (31, 37), "BOOLEAN", "SHARING_AND_UPLOADING"),
]


def enforcement_list():  # type: () -> List[Tuple[str, str, str]]
    groups = {x[1]: x[0] for x in enumerate(_ENFORCEMENT_GROUPS)}
    enforcements = [(x[3], x[0], x[1], x[2]) for x in [*_ENFORCEMENTS, *_COMPOUND_ENFORCEMENTS] if len(x) >= 4 and x[3]]
    enforcements.sort(key=lambda x: (groups[x[0]] if x[0] in groups else 100) * 1000 + (x[2] if isinstance(x[2], int) else next(iter(x[2])) - 0.5))
    return [(x[0].title().replace('_', ' '), x[1].lower(), x[3].lower()) for x in enforcements]


ENFORCEMENTS = {e[0].lower(): e[2].lower() for e in [*_ENFORCEMENTS, *_COMPOUND_ENFORCEMENTS]}

week_days = ('SUNDAY', 'MONDAY', 'TUESDAY', 'WEDNESDAY', 'THURSDAY', 'FRIDAY', 'SATURDAY')
occurrences = ('FIRST', 'SECOND', 'THIRD', 'FOURTH', 'LAST')
months = ('JANUARY', 'FEBRUARY', 'MARCH', 'APRIL', 'MAY', 'JUNE', 'JULY', 'AUGUST', 'SEPTEMBER', 'OCTOBER',
          'NOVEMBER', 'DECEMBER')


def get_cron_week_day(text):   # type: (Optional[str]) -> Optional[int]
    if isinstance(text, str):
        try:
            return week_days.index(text.upper())
        except:
            pass


def get_cron_occurrence(text):  # type: (Optional[str]) -> Optional[int]
    if isinstance(text, str):
        try:
            idx = occurrences.index(text.upper())
            idx += 1
            if idx > 4:
                idx = 4
            return idx
        except:
            pass


def get_cron_month(text):  # type: (Optional[str]) -> Optional[int]
    if isinstance(text, str):
        try:
            m = months.index(text.upper())
            return m + 1
        except:
            pass


def get_cron_month_day(text):  # type: (Optional[str]) -> Optional[int]
    if isinstance(text, str) and text.isnumeric():
        day = int(text)
        if day < 1:
            day = 1
        elif day > 28:
            day = 28
        return day


# OS dependent constants
if sys.platform.startswith('win'):
    OS_WHICH_CMD = 'where'
else:
    OS_WHICH_CMD = 'which'


KEEPER_PUBLIC_HOSTS = {
    'US': 'keepersecurity.com',
    'EU': 'keepersecurity.eu',
    'AU': 'keepersecurity.com.au',
    'CA': 'keepersecurity.ca',
    'JP': 'keepersecurity.jp',
    'GOV': 'govcloud.keepersecurity.us'
}


def get_abbrev_by_host(host):
    # Return abbreviation of the Keeper's public host

    if host.startswith('https:'):
        host = urlparse(host).netloc    # https://keepersecurity.com/api/v2/ --> keepersecurity.com

    keys = [k for k, v in KEEPER_PUBLIC_HOSTS.items() if v == host]
    if keys:
        return keys[0]
    return None


# Messages
# Account Transfer
ACCOUNT_TRANSFER_MSG = """
Your Keeper administrator has enabled the ability to transfer your vault records
in accordance with company operating procedures and policies.
Please acknowledge this change in account settings by typing 'Accept'.
If you do not accept this change by {0}, you will be locked out of your account.
"""

PBKDF2_ITERATIONS = 1_000_000
